# coding: utf-8

# http://www.numpy.org/
import numpy as np

# https://matplotlib.org/
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import cm

# atomman imports
from ..tools import axes_check
from .. import NeighborList

def differential_displacement(system_0, system_1, burgers, plotxaxis='x',
                              plotyaxis='y', xlim=None, ylim=None, zlim=None,
                              neighbors=None, cutoff=None,
                              component='standard', axes=None,
                              plot_scale=1, atom_color=None, atom_cmap=None,
                              display_final_pos=False, return_data=False,
                              matplotlib_axes=None):
    """
    Generates a differential displacement plot for characterizing dislocations.
    
    Parameters
    ----------
    system_0 : atomman.system
        The base/reference system to use.
    system_1 : atomman.system
        The defect/current system to use.
    burgers : array-like object
        The dislocation's Burgers vector.
    plotxaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting x-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'x' = [1, 0, 0].
    plotyaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting y-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'y' = [0, 1, 0].
    xlim : tuple, optional
        The minimum and maximum coordinates along the plotting x-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    ylim : tuple, optional
        The minimum and maximum coordinates along the plotting y-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    zlim : tuple, optional
        The minimum and maximum coordinates normal to the plotting axes
        (i.e. plotxaxis X plotyaxis) to include in the fit.  Values are taken
        in the specified length_unit.  The optimum zlim should encompass only
        a single periodic slice.  If not given, then the limits are set
        based on min and max atomic coordinates along the axis.
    neighbors : atomman.NeighborList, optional
        The neighbor list associated with system_0 to use.  Either neighbors
        or cutoff must be given, or system_0 must have a neighbors attribute.
    cutoff : float, optional
        Cutoff distance for computing a neighbor list for system_0.  Either
        neighbors or cutoff must be given, or system_0 have a neighbors
        attribute.
    component : str, optional
        Indicates the style of the calculation to use.  'standard' (default)
        plots the differential displacements between atoms in the Burgers
        direction.  'xy' plots the differential displacements within the xy
        plotting plane.  This is useful for screw dislocations with localized
        non-screw components.
    axes : array-like object, optional
        3x3 transformation array to apply to the Burgers vector to make it
        correspond to the system's orientation.
    plot_scale : float, optional
        Scalar for multiplying the magnitude of the differential displacement
        arrows.  Default value is 1 (no scaling).
    atom_color : str or list, optional
        Matplotlib color name(s) to use to display the atoms.  If str, that
        color will be assigned to all atypes.  If list, must give a color value
        or None for each atype.  Default value (None) will use cmap instead.
        Note: atom_color and atom_cmap can be used together as long as exactly
        one color or cmap is given for each unique atype.
    atom_cmap : str or list, optional
        Matplotlib colormap name(s) to use to display the atoms.  Atoms will
        be colored based on their initial positions and scaled using zlim. If
        str, that cmap will be assigned to all atypes.  If list, must give a 
        cmap value or None for each atype.  Default value (None) will use 'hsv'
        cmap.  Note: atom_color and atom_cmap can be used together as long as
        exactly one color or cmap is given for each unique atype.
    display_final_pos : bool, optional
        Flag to display positions of atoms and arrows relative to final
        configuration (system_1) rather than initial configuration (system_0).
        Note that this does not affect the atom's cmap color as the initial
        plotzaxis is always used.
    return_data : bool, optional
        If True, will return a dict containing the differential displacement
        vectors and vector positions.  Default is False.  Note: returned values
        are oriented relative to the plotting axes.
    matplotlib_axes : matplotlib.Axes.axes, optional
        An existing matplotlib axes object. If given, the differential displacement
        plot will be added to the specified axes of an existing figure.  This
        allows for subplots to be constructed.

    Returns
    -------
    fig : matplotlib.figure
        The generated figure. Not returned if matplotlib_axes is given.
    data : dict
        Contains differential displacement vectors and arrow plotting information.
        Returned if return_data is True.
    """
    # Transform burgers using axes
    if axes is not None:
        T = axes_check(axes)
        burgers = T.dot(burgers)
    
    # Neighbor list setup
    if neighbors is not None:
        assert cutoff is None, 'neighbors and cutoff cannot both be given'
    elif cutoff is not None:
        neighbors = NeighborList(system=system_0, cutoff=cutoff)
    elif hasattr(system_0, 'neighbors'):
        neighbors = system_0.neighbors
    else:
        raise ValueError('neighbors or cutoff is required')
    
    # Give numeric values for str plot axis terms
    if plotxaxis == 'x':
        plotxaxis = [1.0,0.0,0.0]
    elif plotxaxis == 'y':
        plotxaxis = [0.0,1.0,0.0]
    elif plotxaxis == 'z':
        plotxaxis = [0.0,0.0,1.0]
    
    if plotyaxis == 'x':
        plotyaxis = [1.0,0.0,0.0]
    elif plotyaxis == 'y':
        plotyaxis = [0.0,1.0,0.0]
    elif plotyaxis == 'z':
        plotyaxis = [0.0,0.0,1.0]
    
    # Build transformation matrix, T, from plot axes.
    plotxaxis = np.asarray(plotxaxis, dtype='float64')
    plotyaxis = np.asarray(plotyaxis, dtype='float64')
    T = axes_check([plotxaxis, plotyaxis, np.cross(plotxaxis, plotyaxis)])
    
    # Extract positions and transform using T
    pos_0 = np.inner(system_0.atoms.pos, T)
    pos_1 = np.inner(system_1.atoms.pos, T)
    
    #  Select initial or final reference configurations
    if display_final_pos:
        pos_ref = pos_1
        system_ref = system_1
    else:
        pos_ref = pos_0
        system_ref = system_0

    # Transform burgers using plot axes and separate magnitude and direction
    burgers = T.dot(burgers)
    burgers_mag = np.linalg.norm(burgers)
    burgers_uvect = burgers / burgers_mag
    
    # Set plot limits
    if xlim is None:
        xlim = (pos_ref[:, 0].min(), pos_ref[:, 0].max())
    if ylim is None:
        ylim = (pos_ref[:, 1].min(), pos_ref[:, 1].max())
    if zlim is None:
        zlim = (pos_0[:, 2].min(), pos_0[:, 2].max()) # pylint: disable=invalid-sequence-index
    
    # Identify only atoms in xlim, ylim, zlim
    in_bounds = ((pos_ref[:,0] > xlim[0] - 5.) &
                 (pos_ref[:,0] < xlim[1] + 5.) &
                 (pos_ref[:,1] > ylim[0] - 5.) &
                 (pos_ref[:,1] < ylim[1] + 5.) &
                 (pos_0[:,2] > zlim[0]) & # pylint: disable=invalid-sequence-index
                 (pos_0[:,2] < zlim[1])) # pylint: disable=invalid-sequence-index

    # if matplotlib_axes are passed do not create new
    if matplotlib_axes is not None:
        ax1 = matplotlib_axes
    else:
        # Initial plot setup and parameters
        fig, ax1, = plt.subplots(1, 1, squeeze=True, figsize=(7,7), dpi=72)

    ax1.axis([xlim[0], xlim[1], ylim[0], ylim[1]])
    atom_circle_radius = burgers_mag / 10
    arrow_width_scale = 1. / 200.
    
    # Set default atom_cmap
    if atom_color is None and atom_cmap is None:
        atom_cmap = 'hsv'
    
    # Transform single color/cmap values to lists
    if isinstance(atom_cmap, str):
        if atom_color is not None:
            raise TypeError('atom_cmap and atom_color cannot be str if both are given')
        atom_cmap = [atom_cmap for i in range(system_ref.natypes)]
        atom_color = [None for i in range(system_ref.natypes)]
    elif isinstance(atom_color, str):
        if atom_cmap is not None:
            raise TypeError('atom_cmap and atom_color cannot be str if both are given')
        atom_color = [atom_color for i in range(system_ref.natypes)]
        atom_cmap = [None for i in range(system_ref.natypes)]
    else:
        atom_color = list(atom_color)
        atom_cmap = list(atom_cmap)
    
    # Check atom_color, atom_cmap list compatibility
    if len(atom_cmap) != system_ref.natypes:
        raise ValueError('Invalid number of atom_cmap values')
    if len(atom_color) != system_ref.natypes:
        raise ValueError('Invalid number of atom_color values')
    for ic in range(len(atom_cmap)):
        if atom_cmap[ic] is not None:
            if atom_color[ic] is not None:
                raise ValueError('atom_cmap and atom_color cannot both be given for same atype')
            atom_cmap[ic] = cm.get_cmap(atom_cmap[ic])

    data = {}
    data['centers'] = []
    data['vectors'] = []

    # Loop over all atoms i in plot range
    for i in np.arange(system_ref.natoms)[in_bounds]:
        
        atype = system_ref.atoms.atype[i]
        atype_index = system_ref.atypes.index(atype)
        
        # Plot a circle for atom i
        if atom_cmap[atype_index] is not None:
            color = atom_cmap[atype_index]((pos_0[i, 2] - zlim[0]) / (zlim[1] - zlim[0])) # pylint: disable=invalid-sequence-index
        elif atom_color[atype_index] is not None:
            color = atom_color[atype_index]
        else:
            color = None
        if color is not None:
            ax1.add_patch(mpatches.Circle(pos_ref[i, :2], atom_circle_radius, fc=color, ec='k'))

        # Compute distance vectors between atom i and its neighbors for both systems
        dvectors_0 = np.inner(system_0.dvect(int(i), neighbors[i]), T)
        dvectors_1 = np.inner(system_1.dvect(int(i), neighbors[i]), T)
    
        # Compute differential displacement vectors
        dd_vectors = dvectors_1 - dvectors_0
        
        if display_final_pos:
            # Compute center point positions for the vectors
            arrow_centers = pos_1[i] + dvectors_1 / 2
        else :
            arrow_centers = pos_0[i] + dvectors_0 / 2
        
        if return_data:
            data['vectors'].append(dd_vectors)
            data['centers'].append(arrow_centers)

        # Plot standard differential displacement component
        if component == 'standard':
            # Compute unit distance vectors
            if display_final_pos:
                uvectors = dvectors_1 / np.linalg.norm(dvectors_1, axis=1)[:,np.newaxis]
            else:
                uvectors = dvectors_0 / np.linalg.norm(dvectors_0, axis=1)[:,np.newaxis]
            
            # Compute component of the dd_vector parallel to the burgers vector
            dd_components = dd_vectors.dot(burgers_uvect)
            dd_components[dd_components > burgers_mag / 2] -= burgers_mag
            dd_components[dd_components < -burgers_mag / 2] += burgers_mag
            
            # Scale arrow lengths and vectors
            arrow_lengths = uvectors * dd_components[:,np.newaxis] * plot_scale
            arrow_widths = arrow_width_scale * dd_components * plot_scale
            
            # Plot the arrows
            for center, length, width in zip(arrow_centers, arrow_lengths,
                                             arrow_widths):
                if width > 1e-7:
                    ax1.quiver(center[0], center[1], length[0], length[1],
                               pivot='middle', angles='xy', scale_units='xy',
                               scale=1, width=width)

        
        # Plot xy differential displacement component
        elif component == 'xy':
            # Scale arrow lengths and vectors
            arrow_lengths = dd_vectors[:, :2] * plot_scale
            arrow_lengths[dd_vectors[:, 2] > 0] *= -1
            arrow_widths = arrow_width_scale * (arrow_lengths[:,0]**2
                                              + arrow_lengths[:,1]**2)**0.5
            
            # Plot the arrows
            for center, length, width in zip(arrow_centers, arrow_lengths, arrow_widths):
                if width > 1e-7:
                    ax1.quiver(center[0], center[1], length[0], length[1], width=width,
                               pivot='middle', angles='xy', scale_units='xy', scale=1)
    
    returns = []
    if matplotlib_axes is None:
        returns.append(fig)
    
    if return_data:
        data['centers'] = np.concatenate(data['centers'])
        data['vectors'] = np.concatenate(data['vectors'])
        returns.append(data)

    if len(returns) == 1:
        return returns[0]
    
    elif len(returns) > 1:
        return tuple(returns)


# coding: utf-8
# Standard Python libraries
from pathlib import Path


import pandas as pd


import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/atomman
import atomman as am

import potentials

from .. import CrystalPrototype
from ...tools import aslist, screen_input

@property
def crystal_prototypes(self):
    """list or None: Loaded CrystalPrototype objects"""
    return self.__crystal_prototypes

@property
def crystal_prototypes_df(self):
    """pandas.DataFrame or None: Metadata for loaded CrystalPrototype objects"""
    return self.__crystal_prototypes_df

def _no_load_crystal_prototypes(self):
    """Initializes properties if load_crystal_prototypes is not called"""
    self.__crystal_prototypes = None
    self.__crystal_prototypes_df = None

def load_crystal_prototypes(self, localpath=None, local=None, remote=None, verbose=False):
    """
    Loads crystal prototypes from the database, first checking localpath, then
    trying to download from host.
    
    Parameters
    ----------
    localpath : str, optional
        Path to a local directory to check for records first.  If not given,
        will check localpath value set during object initialization.  If not
        given or set during initialization, then only the remote database will
        be loaded.
    local : bool, optional
        Indicates if records in localpath are to be loaded.  If not given,
        will use the local value set during initialization.
    remote : bool, optional
        Indicates if the records in the remote database are to be loaded.
        Setting this to be False is useful/faster if a local copy of the
        database exists.  If not given, will use the local value set during
        initialization.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    prototypes = {}
    
    # Set localpath, local, and remote as given here or during init
    if localpath is None:
        localpath = self.localpath
    if local is None:
        local = self.local
    if remote is None:
        remote = self.remote
    
    # Check localpath first
    if local is True and localpath is not None:
        for fname in Path(localpath, 'crystal_prototype').glob('*'):
            if fname.suffix in ['.xml', '.json']:
                
                with open(fname, encoding='UTF-8') as f:
                    proto = CrystalPrototype(model=f.read())
                prototypes[proto.id] = proto
                
        if verbose:
            print(f'Loaded {len(prototypes)} local crystal prototypes')
    numlocal = len(prototypes)

    # Load remote
    if remote is True:
        try:
            records = self.cdcs.query(template='crystal_prototype')
        except:
            if verbose:
                print('Failed to load crystal prototypes from remote')
        else:
            if verbose:
                print(f'Loaded {len(records)} remote crystal prototypes')
            
            for i in range(len(records)):
                record = records.iloc[i]
                proto = CrystalPrototype(record.xml_content)
                if proto.id not in prototypes:
                    prototypes[proto.id] = proto

            if verbose and len(prototypes) > 0:
                print(f' - {len(prototypes) - numlocal} new')
    
    # Build crystal_prototypes and crystal_prototypes_df
    if len(prototypes) > 0:
        protos = np.array(list(prototypes.values()))
        protodicts = []
        for proto in prototypes.values():
            protodicts.append(proto.asdict())
        
        self.__crystal_prototypes_df = pd.DataFrame(protodicts).sort_values('id')
        self.__crystal_prototypes = protos[self.crystal_prototypes_df.index]
        self.__crystal_prototypes_df.reset_index(drop=True)
    else:
        if verbose:
            print('No crystal prototypes loaded')
        self.__crystal_prototypes = None
        self.__crystal_prototypes_df = None
        
def get_crystal_prototypes(self, id=None, key=None, name=None, prototype=None,
                   pearson=None, strukturbericht=None, sg_number=None,
                   sg_hm=None, sg_schoenflies=None, keyword=None,
                   verbose=False):
    """
    Get all matching crystal prototypes from the database.
    
    Parameters
    ----------
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    keyword : str, optional
        If given, will limit the search to all records that contain the keyword
        substring.  Cannot be combined with any of the above parameters.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    list of Potential
        The matching potentials
    """

    if keyword is not None:
        try:
            assert id is None
            assert key is None
            assert name is None
            assert prototype is None
            assert pearson is None
            assert strukturbericht is None
            assert sg_hm is None
            assert sg_number is None
            assert sg_schoenflies is None
        except:
            raise ValueError('keyword cannot be combined with the other search limiting paramters')

    # Check loaded values if available
    if self.crystal_prototypes_df is not None:
        
        if keyword is None:

            def strmatch(series, name, val):
                if val is None:
                    return True
                else:
                    return series[name] in aslist(val)

            def intmatch(series, name, val):
                if val is None:
                    return True
                else:
                    val = aslist(val)
                    for i in range(len(val)):
                        val[i] = int(val[i])
                    return series[name] in val
            
            prototypes = self.crystal_prototypes[self.crystal_prototypes_df.apply(strmatch, args=['id', id], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['key', key], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['name', name], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['prototype', prototype], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['pearson', pearson], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['strukturbericht', strukturbericht], axis=1)
                                                &self.crystal_prototypes_df.apply(intmatch, args=['sg_number', sg_number], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['sg_hm', sg_hm], axis=1)
                                                &self.crystal_prototypes_df.apply(strmatch, args=['sg_schoenflies', sg_schoenflies], axis=1)]
        else:
            
            prototypes = []
            for proto in self.crystal_prototypes:
                if keyword.lower() in proto.asmodel().json().lower():
                    prototypes.append(proto)

        if verbose:
            print(len(prototypes), 'matching crystal prototypes found from loaded records')
        return prototypes

    # Check remote values if no loaded values
    else:
        if keyword is None:
            # Build Mongoquery
            mquery = {}

            # Add id query
            if id is not None:
                id = aslist(id)
                mquery['crystal-prototype.id'] = {'$in': id}

            # Add key query
            if key is not None:
                key = aslist(key)
                mquery['crystal-prototype.key'] = {'$in': key}

            # Add name query
            if name is not None:
                name = aslist(name)
                mquery['crystal-prototype.name'] = {'$in': name}

            # Add prototype query
            if prototype is not None:
                prototype = aslist(prototype)
                mquery['crystal-prototype.prototype'] = {'$in': prototype}

            # Add pearson query
            if pearson is not None:
                pearson = aslist(pearson)
                mquery['crystal-prototype.Pearson-symbol'] = {'$in': pearson}

            # Add strukturbericht query
            if strukturbericht is not None:
                strukturbericht = aslist(strukturbericht)
                mquery['crystal-prototype.Strukturbericht'] = {'$in': strukturbericht}

            # Add sg_number query
            if sg_number is not None:
                sg_number = aslist(sg_number)
                for i in range(len(sg_number)):
                    sg_number[i] = int(sg_number[i])
                mquery['crystal-prototype.space-group.number'] = {'$in': sg_number}
            
            # Add sg_hm query
            if sg_hm is not None:
                sg_hm = aslist(sg_hm)
                mquery['crystal-prototype.space-group.Hermann-Maguin'] = {'$in': sg_hm}

            # Add sg_schoenflies query
            if sg_schoenflies is not None:
                sg_schoenflies = aslist(sg_schoenflies)
                mquery['crystal-prototype.space-group.Schoenflies'] = {'$in': sg_schoenflies}

            matches = self.cdcs.query(template='crystal_prototype', mongoquery=mquery)

        else:
            matches = self.cdcs.query(template='crystal_prototype', keyword=keyword)

        if verbose:
            print(len(matches), 'matching crystal prototypes found from remote database')

        if len(matches) > 0:
            matches = matches.sort_values('id').reset_index(drop=True)

            def makeprototypes(series):
                return CrystalPrototype(model=series.xml_content)

            return matches.apply(makeprototypes, axis=1).values
        else:
            return np.array([])

def get_crystal_prototype(self, id=None, key=None, name=None, prototype=None,
                   pearson=None, strukturbericht=None, sg_number=None,
                   sg_hm=None, sg_schoenflies=None, keyword=None,
                   verbose=False):
    
    prototypes = self.get_crystal_prototypes(id=id, key=key, name=name, prototype=prototype,
                   pearson=pearson, strukturbericht=strukturbericht, sg_number=sg_number,
                   sg_hm=sg_hm, sg_schoenflies=sg_schoenflies, keyword=keyword,
                   verbose=verbose)

    # Check number of matches and select
    if len(prototypes) == 1:
        return prototypes[0]
    
    elif len(prototypes) > 1:
        print('Multiple matching crystal prototypes found.')
        for i, prototype in enumerate(prototypes):
            print(i+1, prototype.id)
        choice = int(screen_input('Select which one:'))
        return prototypes[choice-1]
    
    else:
        raise ValueError('No matching prototypes found')

def download_crystal_prototypes(self, localpath=None, prototypes=None, format='json',
                        indent=None, verbose=False):
    """
    Download citation records from the remote and save to localpath.
    
    Parameters
    ----------
    localpath : path-like object, optional
        Path to a local directory where the files will be saved to.  If not
        given, will use the localpath value set during object initialization.
    prototypes : list of CrystalPrototype, optional
        A list of crystal prototypes to download. If not given, all prototypes will
        be downloaded.
    format : str, optional
        The file format to save the record files as.  Allowed values are
        'xml' and 'json' (default).
    indent : int, optional
        The indentation spacing size to use for the locally saved record files.
        If not given, the JSON/XML content will be compact.  Ignored if format
        is 'bib'.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """

    template = 'crystal_prototype'

    # Handle localpath value
    if localpath is None:
        localpath = self.localpath
    if localpath is None:
        raise ValueError('No local path set to save files to')
    
    # Check format value
    format = format.lower()
    allowed_formats = ['xml', 'json']
    if format not in allowed_formats:
        raise ValueError("Format must be 'xml' or 'json'")

    # Create save directory if needed
    save_directory = Path(localpath, template)
    if not save_directory.is_dir():
        save_directory.mkdir(parents=True)

    for fmt in allowed_formats:
        if fmt != format:
            numexisting = len([fname for fname in save_directory.glob(f'*.{fmt}')])
            if numexisting > 0:
                raise ValueError(f'{numexisting} records of format {fmt} already saved locally')

    # Download if needed
    if prototypes is None:
        self.download_records(template=template, localpath=localpath,
                              format=format, indent=indent, verbose=verbose)
        
    else:
        prototypes = aslist(prototypes)

        for prototype in prototypes:
            fname = Path(save_directory, f'{prototype.id}.{format}')
            if format == 'xml':
                with open(fname, 'w', encoding='UTF-8') as f:
                    prototype.asmodel().xml(fp=f, indent=indent)
            elif format == 'json':
                with open(fname, 'w', encoding='UTF-8') as f:
                    prototype.asmodel().json(fp=f, indent=indent)
        if verbose:
            print(f'Copied {len(prototypes)} records to localpath')

def upload_crystal_prototype(self, prototype, workspace=None, verbose=False):
    raise NotImplementedError('To be done...')


# coding: utf-8

# atomman imports
from . import IsotropicVolterraDislocation, Stroh

def solve_volterra_dislocation(C, burgers, ξ_uvw=None, slip_hkl=None,
                               transform=None, axes=None, box=None,
                               m=[1,0,0], n=[0,1,0], tol=1e-8):
    """
    Wrapper function for generating VolterraDislocation children classes
    that provide linear elastic solutions for straight dislocations.
    
    Parameters
    ----------
    C : atomman.ElasticConstants
        The medium's elastic constants.
    burgers : array-like object
        The dislocation's Burgers vector.
    ξ_uvw : array-like object
        The Miller crystal vector associated with the dislocation's line
        direction.  Must be given with slip_hkl to identify the
        transformation matrix to use on C and burgers.
    slip_hkl : array-like object
        The Miller plane indices associated with the dislocation's slip
        plane.  Must be given with ξ_uvw to identify the
        transformation matrix to use on C and burgers.
    transform : array-like object, optional
        A 3x3 set of orthogonal Cartesian vectors that define the
        transformation matrix to use on C and burgers to convert from the
        standard (unit cell) and dislocation orientations.  The 3 vectors
        will automatically be converted into unit vectors.  Using this is
        an alternative to using ξ_uvw and slip_hkl.
    axes : array-like object, optional
        Same as transform.  Retained for backwards compatibility.
    box : atomman.Box, optional
        The unit cell's box that crystal vectors are taken with respect to.
        If not given, will use a cubic box with a=1 meaning that burgers,
        ξ_uvw and slip_hkl will be interpreted as Cartesian vectors.
    m : array-like object, optional
        The m unit vector for the solution.  m, n, and u (dislocation
        line) should be right-hand orthogonal.  Default value is [1,0,0]
        (x-axis).
    n : array-like object, optional
        The n unit vector for the solution.  m, n, and u (dislocation
        line) should be right-hand orthogonal.  Default value is [0,1,0]
        (y-axis). n is normal to the dislocation slip plane.
    tol : float
        Tolerance parameter used to round off near-zero values.  Default
        value is 1e-8.

    Returns
    atomman.defect.VolterraDislocation
        The dislocation solution of the appropriate type.
    """

    try:
        return Stroh(C, burgers, ξ_uvw=ξ_uvw, slip_hkl=slip_hkl, transform=transform,
                   axes=axes, box=box, m=m, n=n, tol=tol)
    except:
        return IsotropicVolterraDislocation(C, burgers, ξ_uvw=ξ_uvw, slip_hkl=slip_hkl,
                                            transform=transform, axes=axes, box=box,
                                            m=m, n=n, tol=tol)
                                            


from copy import deepcopy

import numpy as np
from ..tools import miller
from . import FreeSurface

class StackingFault(FreeSurface):
    """
    Class for generating stacking fault atomic configurations. 
    """
    
    def __init__(self, hkl, ucell, cutboxvector='c', maxindex=None,
                 a1vect_uvw=None, a2vect_uvw=None,
                 conventional_setting='p', tol=1e-8):
        """
        Class initializer.  Identifies the proper rotations for the given hkl plane
        and cutboxvector, and creates the rotated cell.
        
        Parameters
        ----------
        hkl : array-like object
            The free surface plane to generate expressed in either 3 indices
            Miller (hkl) format or 4 indices Miller-Bravais (hkil) format.
        ucell : atomman.System
            The unit cell to use in generating the system.
        cutboxvector : str, optional
            Specifies which of the three box vectors corresponds to the
            out-of-plane vector.  Default value is c.
        maxindex : int, optional
            Max uvw index value to use in identifying the best uvw set for the
            out-of-plane vector.  If not given, will use the largest absolute
            index between the given hkl and the initial in-plane vector guesses.
        a1vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.  If
            not given, will be set to the shortest in-plane lattice vector.
        a2vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.  If
            not given, will be set to the shortest in-plane lattice vector not
            parallel to a1vect_uvw.
        conventional_setting : str, optional
            Allows for rotations of a primitive unit cell to be determined from
            (hkl) indices specified relative to a conventional unit cell.  Allowed
            settings: 'p' for primitive (no conversion), 'f' for face-centered,
            'i' for body-centered, and 'a', 'b', or 'c' for side-centered.  Default
            behavior is to perform no conversion, i.e. take (hkl) relative to the
            given ucell.
        tol : float, optional
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """

        super().__init__(hkl, ucell, cutboxvector=cutboxvector, maxindex=maxindex,
                 conventional_setting=conventional_setting, tol=tol)
        
        # Extract a1vect and a2vect values from uvws and rcell.box
        if a1vect_uvw is None and a2vect_uvw is None:

            # Set a1index and a2index based on cutboxvector
            if self.cutboxvector == 'a':
                a1index, a2index = 1, 2
            elif self.cutboxvector == 'b':
                a1index, a2index = 2, 0
            elif self.cutboxvector == 'c':
                a1index, a2index = 0, 1

            self.a1vect_uvw = self.uvws[a1index]
            self.a2vect_uvw = self.uvws[a2index]
        
        # Set given a1vect_uvw and a2vect_uvw values
        elif a1vect_uvw is not None and a2vect_uvw is not None:
            self.a1vect_uvw = a1vect_uvw
            self.a2vect_uvw = a2vect_uvw
        
        else:
            raise ValueError('a1vect_uvw and a2vect_uvw either both need to be given or not given')
    
    @property
    def a1vect_uvw(self):
        """numpy.ndarray : One of the two conventional lattice shift vectors in Miller or Miller-Bravais indices."""
        return self.__a1vect_uvw
    
    @property
    def a2vect_uvw(self):
        """numpy.ndarray : One of the two conventional lattice shift vectors in Miller or Miller-Bravais indices."""
        return self.__a2vect_uvw
    
    @a1vect_uvw.setter
    def a1vect_uvw(self, value):
        
        value = np.asarray(value)

        # Check shape and convert [uvtw] to [uvw] if needed
        if value.shape == (4,):
            uvw = miller.vector4to3(value)
        elif value.shape == (3,):
            uvw = value
            if self.hkl.shape == (4,):
                value = miller.vector3to4(value)
        else:
            raise ValueError('Invalid uvw shape: must have 3 or 4 values.')

        # Convert to primitive cell
        uvw = miller.vector_conventional_to_primitive(uvw, self.conventional_setting)

        # Convert to Cartesian wrt rotated system
        cart = self.transform.dot(miller.vector_crystal_to_cartesian(uvw, self.ucell.box))

        # Check that Cartesian vector is in slip plane
        if not np.isclose(cart[self.cutindex], 0.0):
            raise ValueError(f'shift vector {value} not in fault plane {self.hkl}')

        # Save uvw and cart
        self.__a1vect_uvw = value
        self.__a1vect_cart = cart

    @a2vect_uvw.setter
    def a2vect_uvw(self, value):
        
        value = np.asarray(value)

        # Check shape and convert [uvtw] to [uvw] if needed
        if value.shape == (4,):
            uvw = miller.vector4to3(value)
        elif value.shape == (3,):
            uvw = value
            if self.hkl.shape == (4,):
                value = miller.vector3to4(value)
        else:
            raise ValueError('Invalid uvw shape: must have 3 or 4 values.')

        # Convert to primitive cell
        uvw = miller.vector_conventional_to_primitive(uvw, self.conventional_setting)

        # Convert to Cartesian wrt rotated system
        cart = self.transform.dot(miller.vector_crystal_to_cartesian(uvw, self.ucell.box))

        # Check that Cartesian vector is in slip plane
        if not np.isclose(cart[self.cutindex], 0.0):
            raise ValueError(f'shift vector {value} not in fault plane {self.hkl}')

        # Save uvw and cart
        self.__a2vect_uvw = value
        self.__a2vect_cart = cart

    @property
    def a1vect_cart(self):
        """numpy.ndarray : One of the two shift vectors in Cartesian relative to system."""
        return self.__a1vect_cart
    
    @property
    def a2vect_cart(self):
        """numpy.ndarray : One of the two shift vectors in Cartesian relative to system."""
        return self.__a2vect_cart
    
    @property
    def faultpos_cart(self):
        """float : The Cartesian position of the slip plane."""
        try:
            return self.__faultpos_cart
        except:
            raise AttributeError('system not yet built. Use build_system() or surface().')
        
    @property
    def faultpos_rel(self):
        """float : The fractional position of the slip plane."""
        try:
            return self.__faultpos_rel
        except:
            raise AttributeError('system not yet built. Use build_system() or surface().')
        
    @property
    def abovefault(self):
        """list : Indices of all atoms in system above the slip plane."""
        try:
            return self.__abovefault
        except:
            raise AttributeError('system not yet built. Use build_system() or surface().')
        
    @faultpos_cart.setter
    def faultpos_cart(self, value):

        # faultpos_rel = (faultpos_cart - origin) / width (for origin, width || cutindex)
        faultpos_rel = ((value - self.system.box.origin[self.cutindex])
                       / self.system.box.vects[self.cutindex, self.cutindex])
        if faultpos_rel < 0.0 or faultpos_rel > 1.0:
            raise ValueError('faultpos is outside system')
        
        self.__faultpos_rel = faultpos_rel
        self.__faultpos_cart = value
        
        # Identify atoms above fault plane position
        self.__abovefault = self.system.atoms.pos[:, self.cutindex] > (self.faultpos_cart)

    @faultpos_rel.setter
    def faultpos_rel(self, value):
        
        if value < 0.0 or value > 1.0:
            raise ValueError('faultpos is outside system')
        
        self.__faultpos_rel = value

        # faultpos_cart = origin + faultpos_rel * width (for origin, width || cutindex) 
        self.__faultpos_cart = (self.system.box.origin[self.cutindex] 
                              + self.faultpos_rel 
                              * self.system.box.vects[self.cutindex, self.cutindex])
        
        # Identify atoms above fault plane position
        self.__abovefault = self.system.atoms.pos[:, self.cutindex] > (self.faultpos_cart)

    def surface(self, shift=None, vacuumwidth=None, minwidth=None, sizemults=None,
                even=False, faultpos_rel=None, faultpos_cart=None):
        """
        Generates the free surface atomic system, which is used as the basis for generating
        the stacking fault configuration(s).
        
        Parameters
        ----------
        shift : array-like object, optional
            Applies a shift to all atoms. Different values allow for free surfaces with
            different termination planes to be selected.
        vacuumwidth : float, optional
            If given, the free surface is created by modifying the system's box to insert
            a region of vacuum with this width. This is typically used for DFT calculations
            where it is computationally preferable to insert a vacuum region and keep all
            dimensions periodic.
        sizemults : list or tuple, optional
            The three System.supersize multipliers [a_mult, b_mult, c_mult] to use on the
            rotated cell to build the final system. Note that the cutboxvector sizemult
            must be an integer and not a tuple.  Default value is [1, 1, 1].
        minwidth : float, optional
            If given, the sizemult along the cutboxvector will be selected such that the
            width of the resulting final system in that direction will be at least this
            value. If both sizemults and minwidth are given, then the larger of the two
            in the cutboxvector direction will be used. 
        even : bool, optional
            A True value means that the sizemult for cutboxvector will be made an even
            number by adding 1 if it is odd.  Default value is False.
        faultpos_rel : float, optional
            The position to place the slip plane within the system given as a
            relative coordinate along the out-of-plane direction.  faultpos_rel
            and faultpos_cart cannot both be given.  Default value is 0.5 if 
            faultpos_cart is also not given.
        faultpos_cart : float, optional
            The position to place the slip plane within the system given as a
            Cartesian coordinate along the out-of-plane direction.  faultpos_rel
            and faultpos_cart cannot both be given.
        """
        
        super().surface(shift=shift, vacuumwidth=vacuumwidth, minwidth=minwidth,
                             sizemults=sizemults, even=even)
        
        # Set Cartesian fault position
        if faultpos_cart is not None:
            if faultpos_rel is not None:
                raise ValueError('faultpos_rel and faultpos_cart cannot both be given')
            self.faultpos_cart = faultpos_cart
        
        # Set relative fault position
        elif faultpos_rel is not None:
            self.faultpos_rel = faultpos_rel
        
        # Set default fault position
        else:
            self.faultpos_rel = 0.5

        return self.system
    
    def fault(self, a1=None, a2=None, outofplane=None, faultshift=None,
              minimum_r=None, a1vect_uvw=None, a2vect_uvw=None,
              faultpos_cart=None, faultpos_rel=None):
        """
        Generates a fault configuration by displacing all atoms above the slip
        plane.

        Parameters
        ----------
        a1 : float, optional
            The fractional coordinate of a1vect to shift by. 
            Default value is 0.0.
        a2 : float, optional
            The fractional coordinate of a2vect to shift by. 
            Default value is 0.0. 
        outofplane : float, optional
            An out-of-plane shift, given in absolute units.
            Default value is 0.0.
        faultshift : array-like object, optional
            The full shifting vector to displace the atoms above the slip
            plane by.  Cannot be given with a1, a2, or outofplane.
        minimum_r : float, optional
            Specifies the minimum allowed interatomic spacing across the slip
            plane.  If any sets of atoms are closer than this value then the
            outofplane shift is increased.  Default value is None, which
            performs no adjustment.
        a1vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.
            Included here for those wishing to override the values set during
            class initialization.
        a2vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.
            Included here for those wishing to override the values set during
            class initialization.
        faultpos_rel : float, optional
            The position to place the slip plane within the system given as a
            relative coordinate along the out-of-plane direction.  Included
            here for those wishing to override the value set when surface()
            was called.  faultpos_rel and faultpos_cart cannot both be given.
        faultpos_cart : float, optional
            The position to place the slip plane within the system given as a
            Cartesian coordinate along the out-of-plane direction.  Included
            here for those wishing to override the value set when surface()
            was called.  faultpos_rel and faultpos_cart cannot both be given.

        Returns
        -------
        atomman.System
            The atomic configuration with stacking fault shift
        """
        # Update uvws and faultpos if given
        if a1vect_uvw is not None:
            self.a1vect_uvw = a1vect_uvw
        if a2vect_uvw is not None:
            self.a2vect_uvw = a2vect_uvw
        if faultpos_cart is not None:
            if faultpos_rel is not None:
                raise ValueError('faultpos_rel and faultpos_cart cannot both be given')
            self.faultpos_cart = faultpos_cart
        elif faultpos_rel is not None:
            self.faultpos_rel = faultpos_rel

        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[self.cutindex] = 1.0
        
        # Identify the two non-cut indices
        inindex = []
        for i in range(3):
            if i != self.cutindex:
                inindex.append(i)
        
        # Calculate faultshift
        if a1 is not None or a2 is not None or outofplane is not None:
            if faultshift is not None:
                raise ValueError('a1, a2, outofplane cannot be given with faultshift')
            if a1 is None:
                a1 = 0.0
            if a2 is None:
                a2 = 0.0
            if outofplane is None:
                outofplane = 0.0
            faultshift = a1 * self.a1vect_cart + a2 * self.a2vect_cart + outofplane * ovect
        
        # Set default faultshift
        elif faultshift is None:
            faultshift = np.array([0.0, 0.0, 0.0])

        # Shift atoms above the fault by faultshift
        sfsystem = deepcopy(self.system)
        sfsystem.atoms.pos[self.abovefault] += faultshift
        sfsystem.wrap()
        
        # Add additional outofplane shift if necessary
        if minimum_r is not None:
            # Get all atoms within minimum_r of fault position
            top_pos = sfsystem.atoms.pos[self.abovefault]
            top_pos = top_pos[top_pos[:, self.cutindex] <= self.faultpos_cart + minimum_r] 
            bot_pos = sfsystem.atoms.pos[~self.abovefault] # 
            bot_pos = bot_pos[bot_pos[:, self.cutindex] >= self.faultpos_cart - minimum_r]
            if top_pos.shape[0] > 0 and bot_pos.shape[0] > 0:
                dmag_min = minimum_r
                dvect_min = None
                
                for i in range(top_pos.shape[0]):
                    dvect = sfsystem.dvect(bot_pos, top_pos[i])
                    if dvect.shape == (3,):
                        dvect = dvect.reshape(1,3)
                    dmag = np.linalg.norm(dvect, axis=1)
                    i = np.argmin(dmag)
                    if dmag[i] < dmag_min:
                        dmag_min = dmag[i]
                        dvect_min = dvect[i]
                
                if dvect_min is not None:
                    
                    new = (minimum_r**2 - dvect_min[inindex[0]]**2 - dvect_min[inindex[1]]**2)**0.5
                    outofplane = new - dvect_min[self.cutindex]
                    faultshift = outofplane * ovect
                    sfsystem.atoms.pos[self.abovefault] += faultshift
                    sfsystem.wrap()
        
        return sfsystem
    
    def iterfaultmap(self, num_a1=None, num_a2=None, outofplane=None,
                     minimum_r=None, a1vect_uvw=None, a2vect_uvw=None,
                     faultpos_cart=None, faultpos_rel=None):
        """
        Iterates over generalized stacking fault configurations associated
        with a 2D map of equally spaced a1, a2 coordinates.

        Parameters
        ----------
        num_a1 : int
            The number of a1 values to generate systems for.  
            Default value is 1 (only generate for a1=0.0).
        num_a2 : int
            The number of a2 values to generate systems for.  
            Default value is 1 (only generate for a2=0.0).
        outofplane : float, optional
            An out-of-plane shift, given in absolute units.
            Default value is 0.0.
        minimum_r : float, optional
            Specifies the minimum allowed interatomic spacing across the slip
            plane.  If any sets of atoms are closer than this value then the
            outofplane shift is increased.  Default value is None, which
            performs no adjustment.
        a1vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.
            Included here for those wishing to override the values set during
            class initialization.
        a2vect_uvw : array-like object, optional
            The crystal vector to use for one of the two shifting vectors.
            Included here for those wishing to override the values set during
            class initialization.
        faultpos_rel : float, optional
            The position to place the slip plane within the system given as a
            relative coordinate along the out-of-plane direction.  Included
            here for those wishing to override the value set when surface()
            was called.  faultpos_rel and faultpos_cart cannot both be given.
        faultpos_cart : float, optional
            The position to place the slip plane within the system given as a
            Cartesian coordinate along the out-of-plane direction.  Included
            here for those wishing to override the value set when surface()
            was called.  faultpos_rel and faultpos_cart cannot both be given.
        
        Yields
        ------
        a1 : float
            The a1 fractional coordinate of a1vect. 
        a2 : float
            The a2 fractional coordinate of a2vect. 
        atomman.System
            The fault configuration associated with the a1, a2 shift.
        """
        # Update uvws and faultpos if given
        if a1vect_uvw is not None:
            self.a1vect_uvw = a1vect_uvw
        if a2vect_uvw is not None:
            self.a2vect_uvw = a2vect_uvw
        if faultpos_cart is not None:
            if faultpos_rel is not None:
                raise ValueError('faultpos_rel and faultpos_cart cannot both be given')
            self.faultpos_cart = faultpos_cart
        elif faultpos_rel is not None:
            self.faultpos_rel = faultpos_rel

        if num_a1 is None:
            num_a1 = 1
        if num_a2 is None:
            num_a2 = 1

        # Construct mesh of regular points
        a1s, a2s = np.meshgrid(np.linspace(0, 1, num_a1, endpoint=False),
                               np.linspace(0, 1, num_a2, endpoint=False))

        for a1, a2 in zip(a1s.flat, a2s.flat):
            yield a1, a2, self.fault(a1=a1, a2=a2, outofplane=outofplane,
                                     minimum_r=minimum_r)


# coding: utf-8

# http://www.numpy.org/
import numpy as np

from . import Shape, Plane

class Cylinder(Shape):
    """
    Class representing a cylinder in space. 
    """
    def __init__(self, center1, center2, radius, endcaps=True):
        """
        Defines a cylinder
        
        Parameters
        ----------
        center1 : array-like object
            A point on the cylinder's axis.  If endcaps is True, the point is taken as the center of
            the cylinder's endcap planes.
        center2 : array-like object
            The center point of one of the flat ends.
        radius : float
            The radius
        endcaps : bool, optional
            Indicates if the cylinder is capped at the two
        """
        self.center1 = center1
        self.center2 = center2
        self.radius = radius
        self.endcaps = endcaps
    
    @property 
    def center1(self):
        """numpy.NDArray : 3D vector position of one cylinder end"""
        return self.__center1
    
    @center1.setter
    def center1(self, value):
        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "center1 must be a 3D vector"
        
        # Save
        self.__center1 = value
    
    @property 
    def center2(self):
        """numpy.NDArray : 3D vector position of one cylinder end"""
        return self.__center2
    
    @center2.setter
    def center2(self, value):
        
        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "center2 must be a 3D vector"
        
        # Save
        self.__center2 = value
    
    @property
    def radius(self):
        """float : the cylinder's radius"""
        return self.__radius

    @radius.setter
    def radius(self, value):
        # Check that value is proper
        value = float(value)
        assert value > 0, "radius must be positive"

        # Save
        self.__radius = value
    
    @property
    def axis(self):
        """numpy.NDArray : 3D unit vector parallel to cylinder axis."""
        d = self.center2 - self.center1
        return d / np.linalg.norm(d)
    
    @property
    def endcaps(self):
        """bool : indicates if the endcaps are included in inside/outside determination"""
        return self.__endcaps

    @endcaps.setter
    def endcaps(self, value):
        if not isinstance(value, bool):
            raise ValueError('endcaps must be bool')
        self.__endcaps = value

    def inside(self, pos, inclusive=True):
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is True.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if inside shape
        """
        
        # Ensure pos is a numpy array 
        pos = np.asarray(pos)
        
        axis = self.axis
        
        # Find points <= radius from axis
        distfromaxis = np.linalg.norm(np.cross(pos - self.center1, axis), axis=-1)
        if inclusive:
            insidecircle = distfromaxis <= self.radius
        else:
            insidecircle = distfromaxis < self.radius

        if self.endcaps:
            # Define end planes
            plane1 = Plane(-axis, self.center1)
            plane2 = Plane(axis, self.center2)
        
            # Combine tests
            return insidecircle & plane1.below(pos, inclusive=inclusive) & plane2.below(pos, inclusive=inclusive)
        
        else:
            return insidecircle


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# atomman imports
from ..core import Atoms, System
import atomman.unitconvert as uc

__all__ = ['point', 'vacancy', 'interstitial', 'substitutional', 'dumbbell']

def point(system, ptd_type='v', pos=None, ptd_id=None,
          db_vect=None, scale=False, atol=None, **kwargs):
    """
    Generates a new System where a point defect has been inserted.
    
    Parameters
    ----------
    system : atomman.System
        The base System to add the defect to.
    ptd_type : str, optional
        Key indicating which type of defect to add. Default value is 'v':
        - 'v' : vacancy.
        - 'i' : positional interstitial.
        - 's' : substitutional.
        - 'db' : dumbbell interstitial.
    pos : array-like object, optional
        Position for adding the defect atom (all styles).
    ptd_id : int, optional
        Atom id where defect is added.  Alternative to using pos
        ('v', 's', 'db' styles).
    db_vect : array-like object, optional
        Vector associated with the dumbbell interstitial ('db' style).
    scale : bool, optional
        Indicates if pos and db_vect are Cartesian (False) or box-relative
        (True). Default value is False.
    atol : float, optional
        Absolute tolerance for position-based searching. Default value is 0.01
        angstroms.
    **kwargs : any, optional
        Keyword arguments corresponding to per-atom property values for the
        new atom ('i', 's', 'db' styles).
    
    Raises
    ------
    AssertionError
        If parameters are given for styles that don't allow them.
    ValueError
        If an invalid ptd_type is given.
    
    Returns
    -------
    atomman.System
        A new system containing the defect.
    """
    
    # If vacancy
    if ptd_type == 'v':
        assert db_vect is None, "db_vect not allowed with ptd_type=='v'"
        assert len(kwargs) == 0, "per-atom properties not allowed with ptd_type=='v'"
        return vacancy(system, pos=pos, ptd_id=ptd_id, scale=scale, atol=atol)
    
    # If interstitial
    elif ptd_type == 'i':
        assert ptd_id is None, "ptd_id not allowed with ptd_type=='i'"
        assert db_vect is None, "db_vect not allowed with ptd_type=='i'"
        return interstitial(system, pos=pos, scale=scale, atol=atol, **kwargs)
    
    # If substitutional
    elif ptd_type == 's':
        assert db_vect is None, "db_vect not allowed with ptd_type=='s'"
        return substitutional(system, pos=pos, ptd_id=ptd_id, scale=scale,
                              atol=atol, **kwargs)
    
    # If dumbbell
    elif ptd_type == 'db':
        return dumbbell(system, pos=pos, ptd_id=ptd_id, db_vect=db_vect,
                        scale=scale, atol=atol, **kwargs)
    
    else:
        raise ValueError('Invalid ptd_type. Options are: v, i, s, or db')

def vacancy(system, pos=None, ptd_id=None, scale=False, atol=None):
    """
    Generates a new system by adding a vacancy point defect.
    1. Removes the indicated atom from the system
    2. Adds per-atom property old_id if it doesn't exist corresponding to the
    atom ids in the original system.
    
    Parameters
    ----------
    system : atomman.System
        The base System to add the defect to.
    pos : array-like object, optional
        Position of the atom to be removed.  Either pos or ptd_id must be
        given.
    ptd_id : int, optional
        Id of the atom to be removed.  Either pos or ptd_id must be given.
    scale : bool, optional
        Indicates if pos is Cartesian (False) or box-relative (True). Default
        value is False.
    atol : float, optional
        Absolute tolerance for position-based searching. Default value is 0.01
        angstroms.
    
    Returns
    -------
    atomman.System
        A new system with the vacancy added.
    """
    # Set default atol
    if atol is None:
        atol = uc.set_in_units(0.01, 'angstrom')
    
    # Identify the id of the atom at pos
    if pos is not None:
        if ptd_id is not None:
            raise ValueError('pos and ptd_id cannot both be supplied')
        
        if scale:
            pos = system.unscale(pos)
        
        dist = np.linalg.norm(system.dvect(pos, system.atoms.pos), axis=1)
        ptd_id = np.where(np.isclose(dist, 0.0, atol=atol))
        if len(ptd_id) == 1 and len(ptd_id[0]) == 1:
            ptd_id = ptd_id[0][0]
        else:
            raise ValueError('Unique atom at pos not identified')
    
    elif ptd_id is not None:
        if ptd_id < 0:
            ptd_id += system.natoms
        if ptd_id < 0 or ptd_id >= system.natoms:
            raise ValueError('invalid ptd_id')
    
    else:
        raise ValueError('Either pos or ptd_id required')
    
    # Generate atomic index list for defect
    index = list(range(system.natoms))
    try:
        index.pop(ptd_id)
    except:
        raise TypeError('ptd_id must be an integer type')
    
    # Build new system
    d_system = System(box=deepcopy(system.box), pbc=deepcopy(system.pbc),
                      atoms=deepcopy(system.atoms[index]),
                      symbols=system.symbols)
    
    # Add property old_id with each atom's original id
    if 'old_id' not in d_system.atoms_prop():
        d_system.atoms.old_id = index
    
    return d_system

def interstitial(system, pos, scale=False, atol=None, **kwargs):
    """
    Generates a new system by adding an interstitial point defect.
    1. Adds a new atom to the end of the Atoms list.
    2. Adds per-atom property old_id if it doesn't exist corresponding to the
    atom ids in the original system.
    3. Sets any of the new atom's per-atom properties to values given as
    kwargs.  Any undefined properties are given zero values except atype,
    which is set to 1.
    
    Parameters
    ----------
    system : atomman.System
        The base System to add the defect to.
    pos : array-like object
        Position of the atom being added.
    scale : bool, optional
        Indicates if pos is Cartesian (False) or box-relative (True).  Default
        value is False.
    atol : float, optional
        Absolute tolerance for position-based searching. Default value is 0.01
        angstroms.
    **kwargs : any, optional
        Keyword arguments corresponding to per-atom property values for the
        new atom.  By default, atype==1 and all other properties are set to
        be all zeros for the property's shape.
    
    Returns
    -------
    atomman.System
        A new system with the interstitial added.
    """
    # Set default atol
    if atol is None:
        atol = uc.set_in_units(0.01, 'angstrom')
    
    if scale:
        pos = system.unscale(pos)
    
    # Check that no atoms are already at pos
    dist = np.linalg.norm(system.dvect(pos, system.atoms.pos), axis=1)
    ptd_id = np.where(np.isclose(dist, 0.0, atol=atol))
    if not (len(ptd_id) == 1 and len(ptd_id[0]) == 0):
        raise ValueError('atom already at pos')
    
    # Generate atomic index list for defect
    index = list(range(system.natoms))
    index.append(0)
    
    # Build new system with atom 0 copied
    d_system = System(box=deepcopy(system.box), pbc=deepcopy(system.pbc),
                      atoms=deepcopy(system.atoms[index]),
                      symbols=system.symbols)
    
    # Add property old_id with each atom's original id
    if 'old_id' not in d_system.atoms_prop():
        d_system.atoms.old_id = index
    
    # Set values for the new atom
    for prop in d_system.atoms_prop():
        if prop == 'atype':
            d_system.atoms.atype[-1] = kwargs.pop('atype', 1)
        elif prop == 'pos':
            d_system.atoms.pos[-1] = pos
        elif prop == 'old_id':
            d_system.atoms.old_id[-1] = kwargs.pop('old_id',
                                                   d_system.atoms.old_id.max()+1)
        else:
            d_system.atoms.view[prop][-1] = kwargs.pop(prop,
                                                       np.zeros_like(d_system.atoms.view[prop][-1]))
    
    return d_system

def substitutional(system, pos=None, ptd_id=None, atype=1,scale=False,
                   atol=None, **kwargs):
    """
    Generates a new system by adding a substitutional point defect.
    1. Moves the indicated atom to the end of the list and changes its atype
    to the value given.
    2. Adds per-atom property old_id if it doesn't exist corresponding to the
    atom ids in the original system.
    3. Sets any of the moved atom's per-atom properties to values given as
    kwargs.  Any undefined properties are left unchanged.
    
    Parameters
    ----------
    system : atomman.System
        The base System to add the defect to.
    pos : array-like object, optional
        Position of the atom being modified.  Either pos or ptd_id must be
        given.
    ptd_id : int, optional
        Id of the atom to be modified.  Either pos or ptd_id must be given.
    atype : int, optional
        Integer atomic type to change the identified atom to.  Must be
        different than the atom's current id.  Default value is 1.
    scale : bool, optional
        Indicates if pos is Cartesian (False) or box-relative (True).  Default
        value is False.
    atol : float, optional
        Absolute tolerance for position-based searching. Default value is 0.01
        angstroms.
    **kwargs : any, optional
        Keyword arguments corresponding to per-atom property values for the
        modified atom.  By default, all properties (except atype) are left
        unchanged.
    
    Returns
    -------
    atomman.System
        A new system with the substitutional added.
    """
    # Set default atol
    if atol is None:
        atol = uc.set_in_units(0.01, 'angstrom')
    
    # Identify the id of the atom at pos
    if pos is not None:
        if ptd_id is not None:
            raise ValueError('pos and ptd_id cannot both be supplied')
        
        if scale:
            pos = system.unscale(pos)
        
        dist = np.linalg.norm(system.dvect(pos, system.atoms.pos), axis=1)
        ptd_id = np.where(np.isclose(dist, 0.0, atol=atol))
        if len(ptd_id) == 1 and len(ptd_id[0]) == 1:
            ptd_id = ptd_id[0][0]
        else:
            raise ValueError('Unique atom at pos not identified')
    
    elif ptd_id is not None:
        if ptd_id < 0:
            ptd_id += system.natoms
        if ptd_id < 0 or ptd_id >= system.natoms:
            raise ValueError('invalid ptd_id')
    
    else:
        raise ValueError('Either pos or ptd_id required')
    
    if system.atoms.atype[ptd_id] == atype:
        raise ValueError('identified atom is already of the specified atype')
    
    # Generate atomic index list for defect
    index = list(range(system.natoms))
    index.pop(ptd_id)
    index.append(ptd_id)
    
    # Build new system
    d_system = System(box=deepcopy(system.box), pbc=deepcopy(system.pbc),
                      atoms=deepcopy(system.atoms[index]),
                      symbols=system.symbols)
    
    # Add property old_id with each atom's original id
    if 'old_id' not in d_system.atoms_prop():
        d_system.atoms.old_id = index
    
    # Set values for the new atom
    for prop in d_system.atoms_prop():
        if prop == 'atype':

            d_system.atoms.atype[-1] = atype
        elif prop == 'pos':
            pass
        else:
            d_system.atoms.view[prop][-1] = kwargs.pop(prop,
                                                       d_system.atoms.view[prop][-1])
    
    return d_system

def dumbbell(system, pos=None, ptd_id=None, db_vect=None, scale=False,
             atol=None, **kwargs):
    r"""
    Generates a new system by adding a dumbbell interstitial point defect.
    
    1. Copies the indicated atom and moves both the original and copy to the
    end of the Atoms list.
    2. Displaces the dumbbell atoms position's by +-db_vect.
    3. Adds per-atom property old_id if it doesn't exist corresponding to the
    atom ids in the original system.
    4. Sets any of the new atom's per-atom properties to values given as
    kwargs.  Any undefined properties are left unchanged.
    
    Parameters
    ----------
    system : atomman.System
        The base System to add the defect to.
    pos : array-like object, optional
        Position of the atom being modified.  Either pos or ptd_id must be
        given.
    ptd_id : int, optional
        Id of the atom to be modified.  Either pos or ptd_id must be given.
    db_vect : array-like object
        Vector shift to apply to the atoms in the dumbbell.
    scale : bool, optional
        Indicates if pos and db_vect are Cartesian (False) or box-relative
        (True).  Default value is False.
    atol : float, optional
        Absolute tolerance for position-based searching. Default value is 0.01
        angstroms.
    \*\*kwargs : any, optional
        Keyword arguments corresponding to per-atom property values for the
        new atom in the dumbbell.  By default, all properties are left
        unchanged (i.e. same as atom that was copied).
    
    Returns
    -------
    atomman.System
        A new system with the dumbbell added.
    """
    # Set default atol
    if atol is None:
        atol = uc.set_in_units(0.01, 'angstrom')
    
    # Identify the id of the atom at pos
    if pos is not None:
        if ptd_id is not None:
            raise ValueError('pos and ptd_id cannot both be supplied')
        
        if scale:
            pos = system.unscale(pos)
        
        dist = np.linalg.norm(system.dvect(pos, system.atoms.pos), axis=1)
        ptd_id = np.where(np.isclose(dist, 0.0, atol=atol))
        if len(ptd_id) == 1 and len(ptd_id[0]) == 1:
            ptd_id = ptd_id[0][0]
        else:
            raise ValueError('Unique atom at pos not identified')
    
    elif ptd_id is not None:
        if ptd_id < 0:
            ptd_id += system.natoms
        if ptd_id < 0 or ptd_id >= system.natoms:
            raise ValueError('invalid ptd_id')
    
    else:
        raise ValueError('Either pos or ptd_id required')
        
    # Unscale db_vect
    if scale:
        db_vect = system.unscale(db_vect)
    
    # Generate atomic index list for defect
    index = list(range(system.natoms))
    index.pop(ptd_id)
    index.append(ptd_id)
    index.append(ptd_id)
    
    # Build new system
    d_system = System(box=deepcopy(system.box), pbc=deepcopy(system.pbc),
                      atoms=deepcopy(system.atoms[index]),
                      symbols=system.symbols)
    
    # Add property old_id with each atom's original id
    if 'old_id' not in d_system.atoms_prop():
        d_system.atoms.old_id = index
    
    # Set values for the new atom
    for prop in d_system.atoms_prop():
        if prop == 'pos':
            d_system.atoms.pos[-2] -= db_vect
            d_system.atoms.pos[-1] += db_vect
        elif prop == 'old_id':
            d_system.atoms.old_id[-1] = kwargs.pop('old_id',
                                                   d_system.atoms.old_id.max()+1)
        else:
            d_system.atoms.view[prop][-1] = kwargs.pop(prop,
                                                       d_system.atoms.view[prop][-1])
    
    return d_system


# coding: utf-8

def dump(system):
    """
    Convert an atomman.System into the cell tuple used by spglib.
    
    Parameters
    ----------
    system : atomman.System
        A atomman representation of a system.
    
    Returns
    -------
    tuple
        (lattice, positions, and numbers) used as inputs for spglib
        functions.
    """
    
    # Get lattice information
    lattice = system.box.vects
    
    # Get atomic positions (in reduced coordinates)
    positions = system.atoms_prop(key='pos', scale=True)
    
    # Get atom numbers (atypes)
    numbers = system.atoms.atype
    
    return (lattice, positions, numbers)


# coding: utf-8

from .Shape import Shape
from .Plane import Plane
from .PlaneSet import PlaneSet
from .Sphere import Sphere
from .Cylinder import Cylinder


__all__ = ['Plane', 'PlaneSet', 'Sphere', 'Cylinder', 'Shape']


# coding: utf-8
# Standard Python libraries
import warnings

# http://www.numpy.org/
import numpy as np

# https://www.scipy.org/
from scipy.optimize import minimize

import matplotlib.pyplot as plt

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ..tools import axes_check
from .GammaSurface import GammaSurface

class SDVPN(object):
    """
    Class representation of the semidiscrete variational Peierls-Nabarro
    dislocation model.
    """
    
    def __init__(self, volterra=None, gamma=None, model=None, 
                 tau=np.zeros((3,3)), alpha=0.0, beta=np.zeros((3,3)),
                 cutofflongrange=None, fullstress=True, cdiffelastic=False,
                 cdiffsurface=True, cdiffstress=False, min_method='Powell',
                 min_kwargs=None, min_options=None):
        """
        Initializes an SDVPN object.
        
        Parameters
        ----------
        volterra : atomman.defect.VolterraDislocation, optional
            The elastic solution for a Volterra dislocation to use as the basis
            of the model. Either volterra or model are required, and both cannot
            be given at the same time.
        gamma : atomman.defect.GammaSurface, optional
            The gamma surface to use for the solution.  Required unless model
            is given and the model content contains gamma surface data.
        model : str or DataModelDict, optional
            Saved data from previous SDVPN runs to load.  Either volterra or
            model are required, and both cannot be given at the same time.
        tau : numpy.ndarray, optional
            A (3,3) array giving the stress tensor to apply to the system
            using the stress energy term.  Only the xy, yy, and yz components
            are used.  Default value is all zeros.
        alpha : list of float, optional
            The alpha coefficient(s) used by the nonlocal energy term.  Default
            value is [0.0].
        beta : numpy.ndarray, optional
            The (3,3) array of beta coefficient(s) used by the surface energy
            term.  Default value is all zeros.
        cutofflongrange : float, optional
            The cutoff distance to use for computing the long-range energy.
            Default value is 1000 angstroms.
        fullstress : bool, optional
            Flag indicating which stress energy algorithm to use.  Default
            value is True.
        cdiffelastic : bool, optional
            Flag indicating if the dislocation density for the elastic energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Default value is False.
        cdiffsurface : bool, optional
            Flag indicating if the dislocation density for the surface energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Default value is True.
        cdiffstress : bool, optional
            Flag indicating if the dislocation density for the stress energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Only matters if fullstress is True.
            Default value is False.
        min_method : str, optional
            The scipy.optimize.minimize method to use.  Default value is
            'Powell'.
        min_kwargs : dict, optional
            Any keyword arguments to pass on to scipy.optimize.minimize besides
            the coordinates, method and options.
        min_options : dict, optional
            Any options to pass on to scipy.optimize.minimize. Default value
            is {}.
        """
        
        # Load solution from existing model
        if model is not None:
            if volterra is not None:
                raise ValueError('model cannot be given with volterra')
            
            self.load(model, gamma=gamma)
        
        # Extract parameters and check solution compatibility
        elif volterra is not None:
            
            # Check that gamma is given
            if gamma is None:
                raise ValueError('gamma is required if volterra is given')
            
            # Check that burgers is in the slip plane
            if not np.isclose(np.dot(volterra.n, volterra.burgers), 0.0):
                raise ValueError('dislocation burgers vector must be in the slip plane')
            
            # Get m, n, ξ, K_tensor, burgers, and transform from volterra
            m, n, ξ = volterra.m, volterra.n, volterra.ξ
            K_tensor = volterra.K_tensor
            burgers = volterra.burgers
            transform = volterra.transform
            
            # Transform K_tensor, burgers and transform to [m,n,ξ] orientation
            mnξ = np.array([m, n, ξ]) # This is transformation matrix to [m,n,ξ] setting
            K_tensor = mnξ.dot(K_tensor.dot(mnξ.T))
            burgers = mnξ.dot(burgers)
            transform = np.matmul(mnξ, transform)
            
            # Check if dislocation system is compatible with gamma surface
            planenormal = transform.dot(gamma.planenormal)
            if not np.isclose(planenormal[0], 0.0) or not np.isclose(planenormal[2], 0.0):
                raise ValueError('different slip planes for gamma and volterra found')

            # Set basic solution definition
            self.__K_tensor = K_tensor
            self.__burgers = burgers
            self.__transform = transform
            self.__gamma = gamma

            # Set options
            self.tau = tau
            self.alpha = alpha
            self.beta = beta
            if cutofflongrange is None:
                self.cutofflongrange = uc.set_in_units(1000, 'angstrom')
            else:
                self.cutofflongrange = cutofflongrange
            self.fullstress = fullstress
            self.cdiffelastic = cdiffelastic
            self.cdiffsurface = cdiffsurface
            self.cdiffstress = cdiffstress
            self.min_method = min_method
            self.min_options = min_options
            self.min_kwargs = min_kwargs

        else:
            raise ValueError('either dislsol or model must be given')
    
    @property
    def x(self):
        """numpy.ndarray : The x coordinates."""
        try:
            return self.__x
        except:
            raise AttributeError('x values not set yet')
    
    @x.setter
    def x(self, value):
        value = np.asarray(value, dtype=float)
        assert value.ndim == 1
        diff = value[1:] - value[:-1]
        assert np.allclose(diff[0], diff), 'x values must be evenly spaced'
        assert diff[0] > 0, 'x values must be in increasing order'
        self.__x = value

    @property
    def disregistry(self):
        """numpy.ndarray : The disregistry vector for each x coordinate."""
        try:
            return self.__disregistry
        except:
            raise AttributeError('disregistry values not set yet')

    @disregistry.setter
    def disregistry(self, value):
        value = np.asarray(value)
        assert value.ndim == 2 and value.shape[1] == 3, 'invalid disregistry dimensions'
        assert np.allclose(value[:,1], 0.0), 'y (i.e. out-of-plane) component of disregistry not supported'
        self.__disregistry = value

    @property
    def K_tensor(self):
        """numpy.ndarray : Dislocation energy coefficient tensor."""
        return self.__K_tensor

    @property
    def burgers(self):
        """numpy.ndarray : Burgers vector."""
        return self.__burgers
        
    @property
    def transform(self):
        """numpy.ndarray : Transformation matrix from standard crystal setting to dislocation solution setting."""
        return self.__transform
    
    @property
    def gamma(self):
        """atomman.defect.GammaSurface : The stacking fault map."""
        return self.__gamma
    
    @property
    def tau(self):
        """numpy.ndarray : The applied 3x3 stress tensor."""
        return self.__tau

    @tau.setter
    def tau(self, value):
        value = np.asarray(value, dtype=float)
        assert value.shape == (3,3)
        self.__tau = value
    
    @property
    def alpha(self):
        """tuple of float : Coefficients for nonlocal energy correction."""
        return self.__alpha
    
    @alpha.setter
    def alpha(self, value):
        try:
            self.__alpha = tuple(value)
        except:
            self.__alpha = (value,)

    @property
    def beta(self):
        """numpy.ndarray : 3x3 coefficients for gradient energy correction."""
        return self.__beta
    
    @beta.setter
    def beta(self, value):
        value = np.asarray(value, dtype=float)
        assert value.shape == (3,3)
        self.__beta = value

    @property
    def cutofflongrange(self):
        """float : Cutoff distance for long-range elastic energy."""
        return self.__cutofflongrange
    
    @cutofflongrange.setter
    def cutofflongrange(self, value):
        self.__cutofflongrange = float(value)

    @property
    def fullstress(self):
        """bool : Flag indicating which stress algorithm was used."""
        return self.__fullstress

    @fullstress.setter
    def fullstress(self, value):
        assert isinstance(value, bool)
        self.__fullstress = value
    
    @property
    def cdiffelastic(self):
        """bool : Flag indicating if elastic energy used central difference for computing the dislocation density."""
        return self.__cdiffelastic
    
    @cdiffelastic.setter
    def cdiffelastic(self, value):
        assert isinstance(value, bool)
        self.__cdiffelastic = value

    @property
    def cdiffsurface(self):
        """bool : Flag indicating if surface energy used central difference for computing the dislocation density."""
        return self.__cdiffsurface
    
    @cdiffsurface.setter
    def cdiffsurface(self, value):
        assert isinstance(value, bool)
        self.__cdiffsurface = value

    @property
    def cdiffstress(self):
        """bool : Flag indicating if stress energy used central difference for computing the dislocation density."""
        return self.__cdiffstress
    
    @cdiffstress.setter
    def cdiffstress(self, value):
        assert isinstance(value, bool)
        self.__cdiffstress = value

    @property
    def min_method(self):
        """str : scipy.optimize.minimize method used."""
        return self.__min_method
    
    @min_method.setter
    def min_method(self, value):
        self.__min_method = str(value)

    @property
    def min_options(self):
        """dict : scipy.optimize.minimize options used."""
        return self.__min_options

    @min_options.setter
    def min_options(self, value):
        if value is None:
            self.__min_options = {}
        elif isinstance(value, dict):
            self.__min_options = value
        else:
            raise TypeError('min_options must be a dict')

    @property
    def min_kwargs(self):
        """dict : scipy.optimize.minimize keywords used."""
        return self.__min_kwargs
    
    @min_kwargs.setter
    def min_kwargs(self, value):
        if value is None:
            self.__min_kwargs = {}
        elif isinstance(value, dict):
            self.__min_kwargs = value
        else:
            raise TypeError('min_kwargs must be a dict')

    @property
    def res(self):
        """OptimizeResult : scipy.optimize.minimize result."""
        try:
            return self.__res
        except:
            return None
    
    def solve(self, x=None, disregistry=None, tau=None, alpha=None, beta=None,
              cutofflongrange=None, fullstress=None, cdiffelastic=None,
              cdiffsurface=None, cdiffstress=None, min_method=None,
              min_kwargs=None, min_options=None):
        """
        Solves the semidiscrete variational Peierls-Nabarro dislocation
        disregistry through energy minimization using the set class
        properties.  All parameters are optional keyword arguments that
        can be used to change any of the previous settings.
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            An array of shape (N) giving the x coordinates corresponding to
            the disregistry solution.
        disregistry : numpy.ndarray, optional
            A (N,3) array giving the initial disregistry vector guess at each
            x coordinate.
        tau : numpy.ndarray, optional
            A (3,3) array giving the stress tensor to apply to the system
            using the stress energy term.  Only the xy, yy, and yz components
            are used.
        alpha : list of float, optional
            The alpha coefficient(s) used by the nonlocal energy term.
        beta : numpy.ndarray, optional
            The (3,3) array of beta coefficient(s) used by the surface energy
            term.
        cutofflongrange : float, optional
            The cutoff distance to use for computing the long-range energy.
        fullstress : bool, optional
            Flag indicating which stress energy algorithm to use.
        cdiffelastic : bool, optional
            Flag indicating if the dislocation density for the elastic energy
            component is computed with central difference (True) or simply
            neighboring values (False).
        cdiffsurface : bool, optional
            Flag indicating if the dislocation density for the surface energy
            component is computed with central difference (True) or simply
            neighboring values (False).
        cdiffstress : bool, optional
            Flag indicating if the dislocation density for the stress energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Only matters if fullstress is True.
        min_method : str, optional
            The scipy.optimize.minimize method to use.
        min_kwargs : dict, optional
            Any keyword arguments to pass on to scipy.optimize.minimize besides
            the coordinates, method and options.
        min_options : dict, optional
            Any options to pass on to scipy.optimize.minimize.
        """

        # Change attribute values if given
        if x is not None:
            self.x = x
        if disregistry is not None:
            self.disregistry = disregistry
        if tau is not None:
            self.tau = tau
        if alpha is not None:
            self.alpha = alpha
        if beta is not None:
            self.beta = beta
        if cutofflongrange is not None:
            self.cutofflongrange = cutofflongrange
        if fullstress is not None:
            self.fullstress = fullstress
        if cdiffelastic is not None:
            self.cdiffelastic = cdiffelastic
        if cdiffsurface is not None:
            self.cdiffsurface = cdiffsurface
        if cdiffstress is not None:
            self.cdiffstress = cdiffstress
        if min_method is not None:
            self.min_method = min_method
        if min_options is not None:
            self.min_options = min_options
        if min_kwargs is not None:
            self.min_kwargs = min_kwargs

        # Check that x and disregistry exist and are of the same length
        if len(self.x) != len(self.disregistry):
            raise ValueError('x and disregistry are not of the same length')
        
        # Define subfunctions
        def decompose(d):
            """Breaks disregistry into components for minimization"""
            d13 = np.concatenate([d[1:-1, 0], d[1:-1, 2]])
            first = d[0]
            last = d[-1]
            return d13, first, last

        def recompose(d13, first, last):
            """Reassembles the disregistry components"""
            half = int(len(d13)/2)
            d = np.zeros((half+2, 3))
            d[0] = first
            d[-1] = last
            d[1:-1, 0] = d13[:half]
            d[1:-1, 2] = d13[half:]
            return d

        def min_func(d13, first, last):
            """Function for minimizing"""
            disregistry = recompose(d13, first, last)
            return self.total_energy(disregistry=disregistry)
        
        # Solve disregistry
        d13, first, last = decompose(self.disregistry)
        res = minimize(min_func, d13, args=(first, last),
                    method=self.min_method, options=self.min_options, **self.min_kwargs)
        self.disregistry = recompose(res.x, first, last)

        self.__res = res
    
    def disldensity(self, x=None, disregistry=None, cdiff=False):
        """
        Computes the dislocation density as the numerical derivative of
        disregistry with respect to x.  Uses either neighboring values
        
            ρ[i] = (δ[i] - δ[i-1]) / (x[i] - x[i-1])
        
        or central difference
        
            ρ[i] = (δ[i+1] - δ[i-1]) / (x[i+1] - x[i-1])
        
        Parameters
        ----------
        cdiff : bool
            Flag indicating how to compute the derivative.  A value of False
            (default) compares nearest values while a value of True compares
            next-nearest (i.e., central difference).
        
        Returns
        -------
        newx : numpy.array
            The x positions corresponding to the dislocation density values.
        rho : numpy.array
            The computed dislocation density.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
        
        # Extract values
        δ = disregistry
        
        if cdiff is False:
            # ρ[i] = (δ[i] - δ[i-1]) / (x[i] - x[i-1])
            ρ = ((δ[1:] - δ[:-1]).T / (x[1:] - x[:-1])).T
            
            # newx is all x except the first
            newx = x[1:]
        
        elif cdiff is True:
            # ρ[i] = (δ[i+1] - δ[i-1]) / (x[i+1] - x[i-1])
            ρ = ((δ[2:] - δ[:-2]).T / (x[2:] - x[:-2])).T
            
            # newx is all x except the first and last
            newx = x[1:-1]
        else:
            raise TypeError('cdiff must be bool')
            
        return (newx, ρ)
    
    def misfit_energy(self, x=None, disregistry=None):
        """
        Computes the misfit energy for the disregistry using the stored gamma
        surface
        
            E_misfit = Σ γ(δ)Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
            
        Returns
        -------
        float
            The misfit energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
        
        # Extract values
        δ = disregistry
        Δx = x[1] - x[0]
        transform = self.transform
        gamma = self.gamma
        
        # Strip out y-component of disregistry and transform for gamma
        disreg = np.vstack([δ[:,0], np.zeros(len(δ)), δ[:,2]]).T
        pos = np.inner(disreg, transform.T)
        
        # Σ γ(δ)Δx
        return Δx * gamma.E_gsf(pos=pos).sum()
    
    def elastic_energy(self, x=None, disregistry=None):
        r"""
        Computes the short-range configuration-dependent elastic energy term
        for the dislocation based on the dislocation density and K_tensor.
        
            E_elastic = 1/(4π) Σ_i Σ_j χ(i,j,Δx) K_lm ρ_l[i] ρ_m[j]
            
            χ(i,j,Δx) = (3/2) Δx² + ψ(i-1,j-1,Δx) + ψ(i,j,Δx) - ψ(i,j-1,Δx) - ψ(j,i-1,Δx)
            
            ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(\|i-j\|Δx)
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The elastic energy for the dislocation.
        """
        # Define subfunctions
        def χ(i, j, Δx):
            """
            Computes the chi subfunction:
                χ(i,j,Δx) = (3/2) Δx² + ψ(i-1,j-1,Δx) + ψ(i,j,Δx)
                                      - ψ(i,j-1,Δx) - ψ(j,i-1,Δx)
            """
            return 3./2. * Δx**2 + (ψ(i-1, j-1, Δx) + ψ(i, j, Δx)
                                    - ψ(i, j-1, Δx) - ψ(j, i-1, Δx))

        def ψ(i, j, Δx):
            """
            Computes the psi subfunction:
                ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(|i-j|Δx)
            """
            # Suppress NaN runtime warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                
                # ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(|i-j|Δx)
                p = 0.5 * (i - j)**2 * Δx**2 * np.log(np.abs(i - j) * Δx)
            
            # Replace NaN values with 0.0
            p[np.isnan(p)] = 0.0
            
            return p
        
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
        
        # Extract values
        δ = disregistry
        Δx = x[1] - x[0]
        cdiff = self.cdiffelastic
        Kij = self.K_tensor
        
        ρ = self.disldensity(x=x, disregistry=δ, cdiff=cdiff)[1]
        j = np.arange(len(ρ), dtype=int)
        
        energy = 0.0
        
        # Compute elastic energy (looping over i, vectorization over j)
        # 1/(4π) Σ_i Σ_j χ(i,j,Δx) K_lm ρ_l[i] ρ_m[j]
        for i in j:
            energy += np.sum( χ(i, j, Δx) * np.inner(ρ[i].dot(Kij), ρ) ) / (4 * np.pi)
        
        return energy
    
    def longrange_energy(self):
        """
        Computes the long-range elastic energy term for the dislocation using
        the K_tensor, Burgers vector and long-range cutoff.  This term is
        configuration-independent and thus the method takes no parameters.
        
            E_longrange = 1/(2π) K_lm b_l b_m ln(L)
        
        Returns
        -------
        float
            The long-range energy for the dislocation.
        """
        # Extract values
        Kij = self.K_tensor
        b = self.burgers
        L = self.cutofflongrange
        
        # Compute long-range energy
        # 1/(2π) K_lm b_l b_m ln(L)
        return np.inner(b.dot(Kij), b) * np.log(L) / (2 * np.pi)
    
    def stress_energy(self, x=None, disregistry=None):
        """
        Computes the stress energy due to the applied stress, tau.
        If fullstress is True, the original stress expression by
        Bulatov and Kaxiras will be used:
        
            E_stress = -1/2 Σ_i (x[i]² - x[i-1]²) ρ_l τ_2l
        
        If fullstress is False, the alternate stress expression by
        Shen and Cheng 10.1016/j.scriptamat.2009.04.047 will be used:
        
            E_stress = -1/2 Σ_i τ_2l (δ_l[i] + δ_l[i+1]) Δx
        
        Note that the Shen and Cheng expression will have a constant
        error associated with it giving an incorrect overall energy, but
        should apply a similar force on the dislocation.

        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The stress energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
        
        # Extract values
        δ = disregistry
        Δx = x[1] - x[0]
        τ = self.tau
        full = self.fullstress
        cdiff = self.cdiffstress
        
        if full is True:
            ρ = self.disldensity(x=x, disregistry=δ, cdiff=cdiff)[1]
            # -1/2 Σ_i (x[i]² - x[i-1]²) ρ_l τ_2l
            return -0.5 * np.sum((x[1:]**2 - x[:-1]**2) * np.inner(ρ, τ[1,:]))
            
        else:
            # Flip sign on tau so energies match full=True
            τ = -τ
            
            # -1/2 Σ_i τ_2l (δ_l[i] + δ_l[i+1]) Δx
            return -0.5 * np.sum(np.inner(τ[1,:], (δ[:-1] + δ[1:]) * Δx))
    
    def surface_energy(self, x=None, disregistry=None):
        """
        Computes the gradient surface energy correction using beta
        coefficients.
        
            E_surface = Σ_j β_lj / 4 Σ_i ρ_l[i]² Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The surface energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
            
        # Extract values
        δ = disregistry
        Δx = x[1] - x[0]
        β = self.beta
        cdiff = self.cdiffsurface
        
        ρ = self.disldensity(x=x, disregistry=δ, cdiff=cdiff)[1]
        
        # Σ_j β_lj / 4 Σ_i ρ_l[i]² Δx
        return np.sum( np.inner(ρ**2 * Δx, β) ) / 4
    
    def nonlocal_energy(self, x=None, disregistry=None):
        """
        Computes the nonlocal energy correction using alpha coefficient(s).
        
            E_nonlocal = Σ_m α_m Σ_i δ[i] (δ[i] - (δ[i+m] + δ[i-m]) / 2) Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The nonlocal energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
            
        # Extract values
        δ = disregistry
        Δx = x[1] - x[0]
        αs = self.alpha
        
        energy = 0.0
        
        # Σ_m α_m Σ_i δ[i] (δ[i] - (δ[i+m] + δ[i-m]) / 2) Δx
        for num, α in enumerate(αs):
            m = num + 1
            dd = δ[m:-m] - 0.5 * (δ[2*m:] + δ[:-2*m])
            energy += α * np.sum(δ[m:-m] * dd * Δx)
            
        return energy
    
    def total_energy(self, x=None, disregistry=None):
        """
        Computes the total energy for the dislocation.
        
            E_total = E_elastic + E_misfit + E_stress + E_surface + E_nonlocal
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The total energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.x
        if disregistry is None:
            disregistry = self.disregistry
            
        return (self.misfit_energy(x, disregistry)
                + self.elastic_energy(x, disregistry)
                + self.longrange_energy()
                + self.stress_energy(x, disregistry)
                + self.nonlocal_energy(x, disregistry)
                + self.surface_energy(x, disregistry))
    
    def check_energies(self, x=None, disregistry=None, 
                       energyperlength_unit='eV/Å'):
        """
        Prints a summary string of all computed energy components.

        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        energyperlength_unit : str, optional
            The units of energy per length to report the dislocation line
            energies in.  Default value is 'eV/Å'.
        """
        hasvals = True
        # Default values are class properties
        if x is None:
            try:
                x = self.x
            except:
                hasvals = False
        if disregistry is None:
            try:
                disregistry = self.disregistry
            except:
                hasvals = False

        if hasvals:
            print(f'Dislocation energy terms in {energyperlength_unit}:')
            print('Misfit energy =    ', uc.get_in_units(self.misfit_energy(x, disregistry), energyperlength_unit))
            print('Elastic energy =   ', uc.get_in_units(self.elastic_energy(x, disregistry), energyperlength_unit))
            print('Long-range energy =', uc.get_in_units(self.longrange_energy(), energyperlength_unit))
            print('Stress energy =    ', uc.get_in_units(self.stress_energy(x, disregistry), energyperlength_unit))
            print('Surface energy =   ', uc.get_in_units(self.surface_energy(x, disregistry), energyperlength_unit))
            print('Nonlocal energy =  ', uc.get_in_units(self.nonlocal_energy(x, disregistry), energyperlength_unit))
            print('Total energy =     ', uc.get_in_units(self.total_energy(x, disregistry), energyperlength_unit))
        else:
            print('x and disregistry must be set/given to check energies')
    
    def disregistry_plot(self, x=None, disregistry=None, figsize=None,
                         length_unit='Å'):
        """
        Creates a simple matplotlib figure showing the disregistry profiles.

        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        figsize : tuple, optional
            matplotlib figure figsize parameter.  Default value is (10, 6).
        length_unit : str, optional
            The unit of length to display positions and disregistry values in.
            Default value is 'Å'.
            
        Returns
        -------
        matplotlib.pyplot.figure
            The generated figure allowing users to perform additional
            modifications.
        """
        hasvals = True
        # Default values are class properties
        if x is None:
            try:
                x = self.x
            except:
                hasvals = False
        if disregistry is None:
            try:
                disregistry = self.disregistry
            except:
                hasvals = False
        if hasvals:

            if figsize is None:
                figsize = (10, 6)

            fig = plt.figure(figsize=figsize)
            
            x = uc.get_in_units(x, length_unit)
            disregistry = uc.get_in_units(disregistry, length_unit)
            plt.plot(x, disregistry[:, 0], label='edge disregistry')
            plt.plot(x, disregistry[:, 1], label='normal disregistry')
            plt.plot(x, disregistry[:, 2], label='screw disregistry')
            plt.legend(fontsize='xx-large')
            plt.xlabel(f'x (${length_unit}$)', size='xx-large')
            plt.ylabel(f'disregistry (${length_unit}$', size='xx-large')
            return fig

        else:
            print('x and disregistry must be set/given to plot')

    def E_gsf_surface_plot(self, x=None, disregistry=None, fmt='ro-',
                           normalize=False, smooth=True, 
                           a1vect=None, a2vect=None, xvect=None,
                           length_unit='Å', energyperarea_unit='eV/Å^2',
                           numx=100, numy=100, figsize=None, **kwargs):
        """
        Extends the GammaSurface.E_gsf_surface_plot() method to plot the
        disregistry path on top of it.
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates. If x
            or disregistry are not set/given, then the disregistry path will
            not be added.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.  If x
            or disregistry are not set/given, then the disregistry path will
            not be added.
        fmt : str, optional
            The matplotlib.pyplot.plot fmt parameter for the disregistry path
            line, i.e. color, marker and line style options.  Default value is
            'ro-': red with circle markers and solid line.
        normalize : bool, optional
            Flag indicating if axes are Cartesian (False, default) or
            normalized by a1, a2 vectors (True).
        smooth : bool, optional
            If True (default), then plot shows smooth interpolated values.
            If False, plot shows nearest raw data values.
        a1vect : np.array, optional
            Crystal vector for the a1 vector to use for plotting.  Default
            value of None uses the saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector to use for plotting.  Default
            value of None uses the saved a2vect.
        xvect : numpy.array, optional
            Crystal vector to align with the plotting x-axis for 
            non-normalized plots.  If not given, this is taken as the Cartesian
            of a1vect.
        length_unit : str, optional
            The unit of length to display non-normalized axes values in.
            Default value is 'Å'.
        energyperarea_unit : str, optional
            The unit of energy per area to display the stacking fault energies
            in. Default value is 'eV/Å^2'.
        numx : int, optional
            The number of plotting points to use along the x-axis.  Default
            value is 100.
        numy : int, optional
            The number of plotting points to use along the y-axis.  Default
            value is 100.       
        figsize : tuple or None, optional
            The figure's x,y dimensions.  If None (default), the values are
            scaled such that the x,y spacings are approximately equal, and the
            larger of the two values is set to 10.
        **kwargs : dict, optional
            Additional keywords are passed into the underlying 
            matplotlib.pyplot.pcolormesh(). This allows control of such things
            like the colormap (cmap).
            
        Returns
        -------
        matplotlib.pyplot.figure
            The generated figure allowing users to perform additional
            modifications.
        """

        # Generate the surface plot
        fig = self.gamma.E_gsf_surface_plot(normalize=normalize, smooth=smooth, 
                                            a1vect=a1vect, a2vect=a2vect, xvect=xvect,
                                            length_unit=length_unit, energyperarea_unit=energyperarea_unit,
                                            numx=numx, numy=numy, figsize=figsize, **kwargs)
        
        hasvals = True
        # Default values are class properties
        if x is None:
            try:
                x = self.x
            except:
                hasvals = False
        if disregistry is None:
            try:
                disregistry = self.disregistry
            except:
                hasvals = False
        if hasvals:

            # Get xvect direction
            if xvect is None:
                if a1vect is None:
                    a1vect = self.gamma.a1vect
                xvect = np.dot(a1vect, self.gamma.box.vects)

            # Transform disregistry to gamma surface pos
            pos = disregistry.dot(self.transform)
            
            # Transform to x, y plotting coordinates and plot
            x, y = self.gamma.pos_to_xy(pos, xvect=xvect)
            x = uc.get_in_units(x, length_unit)
            y = uc.get_in_units(y, length_unit)
            plt.plot(x, y, fmt)

        return fig

    def E_gsf_vs_x_plot(self, x=None, disregistry=None, figsize=None, 
                        length_unit='Å', energyperarea_unit='eV/Å^2'):
        """
        Generates a plot of the stacking fault energy, i.e. misfit energy,
        associated with the disregistry values for each x coordinate.
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates. If x
            or disregistry are not set/given, then the disregistry path will
            not be added.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.  If x
            or disregistry are not set/given, then the disregistry path will
            not be added.
        figsize : tuple or None, optional
            The figure's x,y dimensions.  If None (default), then a figure
            size of (10, 6) will be generated.
        length_unit : str, optional
            The unit of length to display x coordinates in.  Default value is
            'Å'.
        energyperarea_unit : str, optional
            The unit of energy per area to display the stacking fault energies
            in. Default value is 'eV/Å^2'.
            
        Returns
        -------
        matplotlib.pyplot.figure
            The generated figure allowing users to perform additional
            modifications.
        """
        hasvals = True
        # Default values are class properties
        if x is None:
            try:
                x = self.x
            except:
                hasvals = False
        if disregistry is None:
            try:
                disregistry = self.disregistry
            except:
                hasvals = False
        if hasvals:

            if figsize is None:
                figsize=(10,6)

            gsf = self.gamma.E_gsf(pos=disregistry.dot(self.transform))
            
            fig = plt.figure(figsize=figsize)
            plt.plot(uc.get_in_units(x, length_unit),
                     uc.get_in_units(gsf, energyperarea_unit))
            plt.xlabel(f'x-coordinate (${length_unit}$)', size='xx-large')
            plt.ylabel(f'Stacking fault energy (${energyperarea_unit}$)', size='xx-large')
            return fig
        
        else:
            print('x and disregistry must be set/given to check stacking fault energies')


    def load(self, model, gamma=None):
        """
        Load solution from a data model.
        
        Parameters
        ----------
        model : str or DataModelDict
            The semi-discrete-Peierls-Nabarro data model to load.
        gamma : atomman.defect.GammaSurface, optional
            The gamma surface to use.  If not given, will check to see if the
            content is inside model.
        
        Raises
        ------
        ValueError
            If the gamma surface information is not given and it is not found
            in model.
        """
        
        # Identify model element
        try:
            sdpn = DM(model).find('semidiscrete-variational-Peierls-Nabarro')
        except:
            sdpn = DM(model).find('semi-discrete-Peierls-Nabarro')
        
        # Load calculation parameters
        params = sdpn['parameter']
        try:
            self.__transform = uc.value_unit(params['transform'])
        except: 
            self.__transform = uc.value_unit(params['axes'])
        self.__K_tensor = uc.value_unit(params['K_tensor'])
        self.__burgers = uc.value_unit(params['burgers'])
        self.tau = uc.value_unit(params['tau'])
        self.alpha = uc.value_unit(params['alpha'])
        self.beta = uc.value_unit(params['beta'])
        self.cutofflongrange = uc.value_unit(params['cutofflongrange'])
        self.fullstress = params['fullstress']
        self.cdiffelastic = params['cdiffelastic']
        self.cdiffsurface = params['cdiffsurface']
        self.cdiffstress = params['cdiffstress']
        self.min_method = params['min_method']
        self.min_options = params['min_options']
        self.min_kwargs = params.get('min_kwargs', None)
        
        # Load gamma
        if gamma is None:
            try:
                gamma = GammaSurface(model)
            except:
                raise ValueError('No gamma surface in model or given!')
        elif not isinstance(gamma, GammaSurface):
            gamma = GammaSurface(gamma)
        self.__gamma = gamma
        
        # Load calculation solution
        solution = sdpn['solution']
        self.x = uc.value_unit(solution['x'])
        self.disregistry = uc.value_unit(solution['disregistry'])
    
    def model(self, length_unit='Å', energyperarea_unit='eV/Å^2',
              pressure_unit='GPa', include_gamma=False):
        """
        Generate a data model for the object.
        
        Parameters
        ----------
        length_unit : str, optional
            The unit of length to save values as.  Default is 'Å'.
        energyperarea_unit : str, optional
            The unit of energy per area to save fault energy values as.  Only
            used if the gamma surface information is included.  Default value
            is 'eV/Å^2'.
        pressure_unit : str, optional
            The unit of pressure to save values as.  Default is 'GPa'.
        include_gamma : bool, optional
            Flag indicating if the gamma surface data is to be included.
            Default value is False.
            
        Returns
        -------
        DataModelDict
            The data model containing all input parameters and the current
            disregistry vectors.
        """
        model = DM()
        model['semidiscrete-variational-Peierls-Nabarro'] = sdpn = DM()
        
        sdpn['parameter'] = params = DM()
        params['transform'] = uc.model(self.transform, None)
        params['K_tensor'] = uc.model(self.K_tensor, pressure_unit)
        params['tau'] = uc.model(self.tau, pressure_unit)
        params['alpha'] = uc.model(self.alpha, pressure_unit+'/'+length_unit)
        
        params['beta'] = uc.model(self.beta, pressure_unit+'*'+length_unit)
        params['cdiffelastic'] = self.cdiffelastic
        params['cdiffsurface'] = self.cdiffsurface
        params['cdiffstress'] = self.cdiffstress
        params['cutofflongrange'] = uc.model(self.cutofflongrange, length_unit)
        params['burgers'] = uc.model(self.burgers, length_unit)
        params['fullstress'] = self.fullstress
        params['min_method'] = self.min_method
        params['min_options'] = self.min_options
        if len(self.min_kwargs) > 0:
            params['min_kwargs'] = self.min_kwargs
        
        if include_gamma is True:
            sdpn['generalized-stacking-fault'] = self.gamma.model(
                                                  length_unit=length_unit,
                                                  energyperarea_unit=energyperarea_unit)
        
        sdpn['solution'] = solution = DM()
        solution['x'] = uc.model(self.x, length_unit)
        solution['disregistry'] = uc.model(self.disregistry, length_unit)
        
        return model


# coding: utf-8
# Standard Python libraries
import ast

# https://pypi.python.org/pypi/numericalunits
import numericalunits as nu

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# http://www.numpy.org/
import numpy as np

def build_unit():
    """
    Saves numericalunits attributes to global dictionary unit so the values
    can be retrieved by their string names.
    """
    
    # Define global dictionary
    global unit
    unit = {}
    
    # Copy all float attributes of numericalunits to unit
    for key, value in nu.__dict__.items():
        try:
            key = key.decode('UTF-8')
        except:
            pass
        if key[:1] != '_' and isinstance(value, float):
            unit[key] = value

def reset_units(seed=None, **kwargs):
    """
    Extends numericalunits.reset_units() by allowing for working units to be
    defined.  If no working units are specified, then random working units are
    used just like the default numericalunits behavior.  Otherwise, use the
    specified working units and SI.
    
    Parameters
    ----------
    seed : int, optional
        random number seed to use in generating random working units.
        seed='SI' will use SI units.  Cannot be given with the other
        parameters.
    length : str, optional
        Unit of length to use for the working units.
    mass : str, optional
        Unit of mass to use for the working units.
    time : str, optional
        Unit of time to use for the working units.
    energy : str, optional
        Unit of energy to use for the working units.
    charge : str, optional
        Unit of charge to use for the working units.
        
    Raises
    ------
    ValueError
        If seed is given with any other parameters, or if more than four of
        the working unit parameters are given.
    """
    
    # Generate random base working units
    if (len(kwargs) == 0):
        
        nu.reset_units(seed)
        build_unit()
    
    # Generate SI + defined working units
    elif seed is None:
        
        # Check that no more than 4 working units are defined
        if len(kwargs) > 4:
            raise ValueError('Only four working units can be defined')
        
        # Set base units to 1 (working units to SI)
        nu.reset_units('SI')
        build_unit()
        
        # Scale base units by working units
        if 'length' in kwargs:
            nu.m = unit['m'] / unit[kwargs['length']]
        
        if 'mass' in kwargs:
            nu.kg = unit['kg'] / unit[kwargs['mass']]
        
        if 'time' in kwargs:
            nu.s = unit['s'] / unit[kwargs['time']]
        
        if 'charge' in kwargs:
            nu.C = unit['C'] / unit[kwargs['charge']]
        
        # Scale derived units by working units
        if 'energy' in kwargs:
            J = unit['J'] / unit[kwargs['energy']]
            
            # Scale base units by derived units
            if 'mass' not in kwargs:
                nu.kg = J * nu.s**2 / nu.m**2
            elif 'time' not in kwargs:
                nu.s = (nu.kg * nu.m**2 / J)**0.5
            elif 'length' not in kwargs:
                nu.m = (J * nu.s**2 / nu.kg)
        
        # Rebuild derived units and unit dictionary
        nu.set_derived_units_and_constants()
        build_unit()
        
    else:
        raise ValueError('seed cannot be given with any other parameters')

def set_literal(term):
    """
    Convert string 'value unit' to numbers in working units.
    
    Parameters
    ----------
    term : str
        String containing value and associated unit. If unit is not given,
        then the value is converted to a float and assumed to be in working
        units.
        
    Returns
    -------
    numpy.ndarray
        The numerical value of term in working units.
        
    Raises
    ------
    ValueError
        If no valid float value can be parsed.
    """
    
    # Set splitting point j to end of term (i.e. assume no units given)
    j = len(term)
    
    # Loop until done
    while True:
        
        # Split term into value, unit terms
        value = term[:j].strip()
        unit = term[j:].strip()
        if len(unit) == 0:
            unit = None

        # Return number if value, unit pair is valid
        try: 
            return set_in_units(ast.literal_eval(value), unit)
        except: 
            # Find the next splitting point
            try: 
                j = term[:j].rindex(' ')
            except: 
                raise ValueError('Failed to parse term')

def set_in_units(value, units):
    """
    Convert value from specified units to working units.
    
    Parameters
    ----------
    value : int, float, numpy.ndarray, etc.
        A numerical value or list/array of values.
    units : str
        The units that value is in.
        
    Returns
    -------
    numpy.ndarray
        The given value converted from the specified units to working units.
    """
    units = parse(units)
    return np.asarray(value) * units

def get_in_units(value, units):
    """
    Convert value from working units to specified units.
    
    Parameters
    ----------
    value : int, float, numpy.ndarray, etc.
        A numerical value or list/array of values.
    units : str
        The units to convert value to (from working units).
        
    Returns
    -------
    numpy.ndarray
        The given value converted to the specified units from working units.
    """
    units = parse(units)
    return np.asarray(value) / units

def value_unit(term):
    """
    Reads numerical value from dictionary containing 'value' and 'unit' keys.
    
    Parameters
    ----------
    term : dict
        Dictionary containing 'value' and 'unit' keys.
        
    Returns
    -------
    numpy.ndarray
        The result of calling set_in_units() by passing the dictionary keys 
        'value' and 'unit' as parameters.
    
    """
    unit = term.get('unit', None)
    if unit is None:
        value = np.asarray(term['value'])
    else:
        value = set_in_units(term['value'], unit)
    
    if 'shape' in term:
        shape = tuple(term['shape'])
        value = value.reshape(shape)
    
    return value
    
def error_unit(term):
    """
    Reads numerical error from dictionary containing 'error' and 'unit' keys.
    
    Parameters
    ----------
    term : dict
        Dictionary containing 'error' and 'unit' keys.
        
    Returns
    -------
    numpy.ndarray
        The result of calling set_in_units() by passing the dictionary keys 
        'error' and 'unit' as parameters.
    
    """
    unit = term.get('unit', None)
    if unit is None:
        error = np.asarray(term['error'])
    else:
        error = set_in_units(term['error'], unit)
    
    if 'shape' in term:
        shape = tuple(term['shape'])
        error = error.reshape(shape)
    
    return error
    
def model(value, units=None, error=None):
    """
    Generates DataModelDict representation of data.
    
    Parameters
    ----------
    value : float, numpy.ndarray, etc.
        A numerical value or list/array of values.
    units : str
        The units to convert value to (from working units).
    error : float, numpy.ndarray, etc., optional
        A value error to include.  If given, must be the same
        size/shape as value.
    
    Returns
    -------
    DataModelDict
        Model representation of the value(s).
    """
    
    datamodel = DM()
    
    if units is not None:
        value = get_in_units(value, units)
    
    if error is not None:
        error = get_in_units(error, units)
    
    # Single value
    if value.ndim == 0:
        datamodel['value'] = value
        if error is not None:
            datamodel['error'] = error
    
    # 1D array
    elif value.ndim == 1:
        datamodel['value'] = value.tolist()
        if error is not None:
            datamodel['error'] = error.tolist()
            
    # Higher-order array requires shape
    else:
        shape = value.shape
        datamodel['value'] = value.flatten().tolist()
        if error is not None:
            datamodel['error'] = error.flatten().tolist()
        datamodel['shape'] = list(shape)
    
    if units is not None:
        datamodel['unit'] = units
    return datamodel

def parse(units):
    """
    Convert units as strings (or None) into scaling numbers.  This function
    allows for complex unit definitions with operators:
    
    - '()' for defining order of operations
    - '*' for multiplication.
    - '/' for division.
    - '^' for powers.
    
    Parameters
    ----------
    units : str or None
        String consisting of defined unit names, operators, and numerical 
        values to interpret.
        
    Returns
    -------
    float
        The scaling factor for converting numbers in the given units to
        working units. If units is None or == 'scaled', then this value is
        1.0.
    """
    
    # Units of None does no scaling
    if units is None or units == 'scaled':
        return 1

    # Parse string and return number value
    elif isinstance(units, str):
        i = 0
        terms = []
        
        # Break into terms
        while i < len(units):
            
            # parse terms in parentheses first
            if units[i] == '(':
                j = i+1
                pcount = 0
                while True:
                    if j == len(units):
                        raise ValueError('Invalid () terms.')
                    elif units[j] == ')':
                        if pcount == 0:
                            break
                        else:
                            pcount -= 1
                    elif units[j] == '(':
                        pcount += 1
                    j += 1
                    
                terms.append(parse(units[i+1:j]))
                i = j+1
            
            # append string terms
            elif units[i].isalpha():
                term = ''
                while i < len(units) and units[i] not in ' */^\n\r\t':
                    term += units[i]
                    i += 1
                terms.append(unit[term])
            
            # append numeric terms
            elif units[i].isdigit() or units[i] == '-' or units[i] == '.':
                term = ''
                while i < len(units) and units[i] not in ' */^\n\r\t':
                    term += units[i]
                    i += 1
                terms.append(float(term))
            
            # append operators
            elif units[i] in '*/^':
                terms.append(units[i])
                i += 1
            
            # ignore excess white characters
            elif units[i] in ' \n\r\t':
                i += 1
            
            # issue error for unmatched ) parentheses
            elif units[i] == ')':
                raise ValueError('Invalid () terms.')
            
            else:
                raise ValueError('Unknown character: %s' % units[i])
        
        # Compute powers
        while '^' in terms:
            c = terms.index('^')
            value = [terms[c-1] ** terms[c+1]]
            terms = terms[:c-1] + value + terms[c+2:]
        
        # Compute multiplication and division
        while len(terms) > 1:
            if terms[1] == '*':
                value = [terms[0] * terms[2]]
                terms = value + terms[3:]
            elif terms[1] == '/':
                value = [terms[0] / terms[2]]
                terms = value + terms[3:]
            else:
                raise ValueError('Invalid string format')
        
        return terms[0]

    # Else assume units is already a number
    else:
        return units

# Initial reset.  Only called first time module is loaded.
reset_units()


# coding: utf-8
# Standard Python libraries
import os
import glob

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

from .Log import Log

class NEBLog(object):
    
    def __init__(self, neblog='log.lammps', replicalogs='log.lammps.*',
                 rootdir=None):
        """
        Initializes reader for LAMMPS NEB calculation log files.
        
        Parameters
        ----------
        neblog : str, optional
            The filename for the primary LAMMPS NEB log file. Default value is
            'log.lammps'.
        replicalogs : str, optional
            The filename with wildcard for the LAMMPS NEB replica log files.
            Default value is 'log.lammps.*'
        rootdir : str, optional
            The root directory to use with respect to the neblog and
            replicalogs parameters.  Default value is None, which will assume
            the current working directory for relative paths.        
        """
        
        self.load(neblog=neblog, replicalogs=replicalogs,  rootdir=rootdir)

    @property
    def nreplicas(self):
        """int: The number of replicas"""
        return self.__nreplicas
    
    @property
    def minrun(self):
        """pandas.DataFrame: The NEB log data for the minimization steps."""
        return self.__minrun
    
    @property
    def climbrun(self):
        """pandas.DataFrame: The NEB log data for the barrier climb steps."""
        return self.__climbrun
    
    @property
    def logs(self):
        """list of atomman.lammps.log: The LAMMPS log files for each replica."""
        return self.__logs
    
    def load(self, neblog='log.lammps', replicalogs='log.lammps.*',
             rootdir=None):
        """
        Loads LAMMPS NEB calculation log file data.
        
        Parameters
        ----------
        neblog : str, optional
            The filename for the primary LAMMPS NEB log file. Default value is
            'log.lammps'.
        replicalogs : str, optional
            The filename with wildcard for the LAMMPS NEB replica log files.
            Default value is 'log.lammps.*'
        rootdir : str, optional
            The root directory to use with respect to the neblog and
            replicalogs parameters.  Default value is None, which will assume
            the current working directory for relative paths.        
        """
        if rootdir is not None:
            neblog = os.path.join(rootdir, neblog)
            replicalogs = os.path.join(rootdir, replicalogs)
        replicalogs = replicalogs.replace('*', '%i')

        # First pass
        with open(neblog) as f:
            lines = f.readlines()

            # Count terms to identify number of replicas
            terms = lines[3].split()
            self.__nreplicas = int((len(terms) - 9) / 2)
            
            column_names = 'Step MaxReplicaForce MaxAtomForce GradV0 GradV1 GradVc EBF EBR RDT'.split()
            for i in range(self.nreplicas):
                column_names.append('RD%i' %(i+1))
                column_names.append('PE%i' %(i+1))

            climb_start = None
            nrows = None
            for i in range(3, len(lines)):
                if lines[i][:4] == 'Step':
                    climb_start = i + 1
                    nrows = i - 4

        # Second pass
        self.__minrun = pd.read_csv(neblog, names=column_names, skiprows=3, nrows=nrows, delim_whitespace=True, skip_blank_lines=True)
        self.__climbrun = pd.read_csv(neblog, names=column_names, skiprows=climb_start, delim_whitespace=True, skip_blank_lines=True)
        
        self.__logs = []
        for i in range(self.nreplicas):
            self.__logs.append(Log(replicalogs % (i) ))
       
    def get_neb_path(self, step):
        """
        Retrieves the reaction coordinates and corresponding potential
        energies for a given simulation step.
        
        Parameters
        ----------
        step : int
            The run step to retrieve values for.
            
        Returns
        -------
        reaction_coordinates : numpy.ndarray
            The reaction coordinates
        potential_energies : numpy.ndarray
            The potential energies
        """
        reaction_coordinate = []
        potential_energy = []
        joined = pd.concat([self.minrun, self.climbrun])
        match = joined[joined.Step == step]
        if len(match) > 0:
            match = match.iloc[0]
        else:
            raise ValueError('Step value not found in NEB log file')
        key = 'RD%i'
        for i in range(1, self.nreplicas+1):
            reaction_coordinate.append(match[key%i])    

        for replica in range(self.nreplicas):
            match = None    
            for sim in self.logs[replica].simulations:
                try:
                    match = sim['thermo'][sim['thermo'].Step == step].iloc[0]
                except:
                    continue
                else:
                    break
            if match is None:
                raise ValueError('Step value not found in log.lammps.%i file' %replica)
            potential_energy.append(match.PotEng)

        reaction_coordinates = np.array(reaction_coordinate)
        potential_energies = np.array(potential_energy) - potential_energy[0]

        return reaction_coordinates, potential_energies

    def get_barrier(self, reverse=False):
        """
        Returns the barrier energy calculated from the final NEB simulation step.
        
        Parameters
        ----------
        reverse : bool, optional
            Indicates if the energy barrier returned is the forward barrier
            relative to the first replica (False, default), or is the reverse
            barrier relative to the last replica (True).
            
        Returns
        -------
        float
            The energy barrier relative to one of the endpoint replicas.
        """
        potential_energies = self.get_neb_path(self.climbrun.Step.values[-1])[1]
        
        if reverse is False:
            reference = potential_energies[0]
        elif reverse is True:
            reference = potential_energies[-1]
        else:
            raise ValueError('Parameter reverse must be bool')
        
        return np.max(potential_energies) - reference


# coding: utf-8
 
# atomman imports
from .. import load_system_model
from ...library import Database
from ...tools import identifyfamily, screen_input

def load(id=None, key=None, name=None, prototype=None,
         pearson=None, strukturbericht=None, sg_number=None,
         sg_hm=None, sg_schoenflies=None, keyword=None,
         a=None, b=None, c=None, alpha=None, beta=None, gamma=None, symbols=None,
         database=None, localpath=None, local=True, remote=True, verbose=False):
    """
    Loads a crystal prototype record from the library. If multiple matches
    are found based on inputs a selection menu will appear.
    
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optionypeal
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    keyword : str, optional
        If given, will limit the search to all records that contain the keyword
        substring.  Cannot be combined with any of the above parameters.
    a : float, optional
        The a lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    b : float, optional
        The b lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    c : float, optional
        The c lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    alpha : float, optional
        The alpha lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    beta : float, optional
        The beta lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    gamma : gamma, optional
        The alpha lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
        Useful if the symbols for the model differ from the standard element
        tags or if the poscar file has no elemental information.
    database : atomman.library.Database, optional
        A pre-defined Database object to use.  If not given, will initialize
        a new Database object.  Passing in a database can save time if multiple
        calls are made for the same record type. 
    localpath : str, optional
        The local library path to use when initializing a new Database.  IF not
        given, will use the default localpath.  Ignored if database is given. 
    local : bool, optional
        Indicates if the Database object is to look for local records.  Default
        is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the Database object is to look for remote records.  Default
        is True.  Ignored if database is given.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    system : atomman.System
        The system object generated from the crystal prototype.
    """

    # Create Database object and load if needed
    if database is None:
        if local is True:
            database = Database(load='crystal_prototype', localpath=localpath,
                                local=local, remote=remote, verbose=verbose)
        else:
            database = Database(local=local, remote=remote, verbose=verbose)
    
    # Get crystal prototype record
    prototype = database.get_crystal_prototype(id=id, key=key, name=name, prototype=prototype,
         pearson=pearson, strukturbericht=strukturbericht, sg_number=sg_number,
         sg_hm=sg_hm, sg_schoenflies=sg_schoenflies, keyword=keyword, verbose=verbose)
    
    # Retrieve unit cell information and set symbols
    ucell = prototype.ucell
    ucell.symbols = symbols
    
    # Allow lattice constants to be set based on crystal family
    family = identifyfamily(ucell.box)
    if family == 'cubic':
        if a is not None:
            if b is not None or c is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a can be set for cubic prototypes')
            ucell.box_set(a=a, b=a, c=a, scale=True)
            
    elif family == 'hexagonal':
        if a is not None and c is not None:
            if b is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, c can be set for hexagonal prototypes')
            ucell.box_set(a=a, b=a, c=c, gamma=120, scale=True)
        elif a is not None or c is not None:
            raise ValueError('All or neither of a, c must be set for hexagonal prototypes')
            
    elif family == 'tetragonal':
        if a is not None and c is not None:
            if b is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, c can be set for tetragonal prototypes')
            ucell.box_set(a=a, b=a, c=c, scale=True)
        elif a is not None or c is not None:
            raise ValueError('All or neither of a, c must be set for tetragonal prototypes')
            
    elif family == 'rhombohedral':
        if a is not None and alpha is not None:
            if b is not None or c is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, alpha can be set for rhombohedral prototypes')
            ucell.box_set(a=a, b=a, c=a, alpha=alpha, beta=alpha, gamma=alpha, scale=True)
        elif a is not None or alpha is not None:
            raise ValueError('All or neither of a, alpha must be set for rhombohedral prototypes')
            
    elif family == 'orthorhombic':
        if a is not None and b is not None and c is not None:
            if alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, b, c can be set for orthorhombic prototypes')
            ucell.box_set(a=a, b=b, c=c, scale=True)
        elif a is not None or b is not None or c is not None:
            raise ValueError('All or neither of a, b, c must be set for orthorhombic prototypes')
            
    elif family == 'monoclinic':
        if a is not None and b is not None and c is not None and beta is not None:
            if alpha is not None or gamma is not None:
                raise ValueError('Only a, b, c, beta can be set for monoclinic prototypes')
            ucell.box_set(a=a, b=b, c=c, beta=beta, scale=True)
        elif a is not None or b is not None or c is not None or beta is not None:
            raise ValueError('All or neither of a, b, c, beta must be set for monoclinic prototypes')
            
    else:
        if a is not None and b is not None and c is not None and alpha is not None and beta is not None and gamma is not None:
            ucell.box_set(a=a, b=b, c=c, alpha=alpha, beta=beta, gamma=gamma, scale=True)
        elif a is not None or b is not None or c is not None or alpha is not None or beta is not None or gamma is not None:
            raise ValueError('All or neither of a, b, c, alpha, beta, gamma must be set for triclinic prototypes')
    
    return ucell


# coding: utf-8
# Standard Python libraries
import warnings

# http://www.numpy.org/
import numpy as np

# atomman imports
from . import VolterraDislocation
from ..tools import vect_angle, axes_check

class IsotropicVolterraDislocation(VolterraDislocation):
    """
    Class for representing the isotropic Volterra solution for a straight dislocation.
    """
    
    def solve(self, C, burgers, ξ_uvw=None, slip_hkl=None, transform=None,
              axes=None, box=None, m=[1,0,0], n=[0,1,0], tol=1e-8):
        """
        Computes the elastic solution for an isotropic volterra dislocation.
        
        Parameters
        ----------
        C : atomman.ElasticConstants
            The medium's elastic constants.
        burgers : array-like object
            The dislocation's Burgers vector.
        ξ_uvw : array-like object
            The Miller crystal vector associated with the dislocation's line
            direction.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        slip_hkl : array-like object
            The Miller plane indices associated with the dislocation's slip
            plane.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        transform : array-like object, optional
            A 3x3 set of orthogonal Cartesian vectors that define the
            transformation matrix to use on C and burgers to convert from the
            standard (unit cell) and dislocation orientations.  The 3 vectors
            will automatically be converted into unit vectors.  Using this is
            an alternative to using ξ_uvw and slip_hkl.
        axes : array-like object, optional
            Same as transform.  Retained for backwards compatibility.
        box : atomman.Box, optional
            The unit cell's box that crystal vectors are taken with respect to.
            If not given, will use a cubic box with a=1 meaning that burgers,
            ξ_uvw and slip_hkl will be interpreted as Cartesian vectors.
        m : array-like object, optional
            The m unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [1,0,0]
            (x-axis).
        n : array-like object, optional
            The n unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [0,1,0]
            (y-axis). n is normal to the dislocation slip plane.
        tol : float
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """
        VolterraDislocation.solve(self, C, burgers, ξ_uvw=ξ_uvw, slip_hkl=slip_hkl,
                                  transform=transform, axes=axes, box=box,
                                  m=m, n=n, tol=tol)
        
        # Check that C is isotropic
        if not C.is_normal('isotropic', atol=0.0, rtol=1e-4):
            raise ValueError('C must be isotropic elastic constants')
        self.__C = C.normalized_as('isotropic')
        
        # Save shear modulus and Poissons ratio
        bulk = self.C.bulk()
        self.__mu = self.C.shear()
        self.__nu = (3 * bulk - 2 * self.mu) / (2 * (3 * bulk + self.mu))
    
    @property
    def mu(self):
        return self.__mu
    
    @property
    def nu(self):
        return self.__nu
    
    @property
    def K_tensor(self):
        
        # Construct K_tensor in standard setting
        K_e = self.mu / (1 - self.nu)
        K_s = self.mu
        K = np.array([[K_e, 0.0, 0.0],
                      [0.0, K_e, 0.0],
                      [0.0, 0.0, K_s]])

        # Transform tensor from m, n, ξ system
        trans = np.array([self.m, self.n, self.ξ])
        K = trans.T.dot(K.dot(trans))
        
        # Round away near-zero terms
        K[np.isclose(K / K.max(), 0.0, atol=self.tol)] = 0.0
        return K

    def theta(self, pos):
        """
        Computes arctan(y / x) ranging from -π to π.
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
            
        Returns
        -------
        numpy.ndarray
            The theta angles for each 3D position.
        """
        pos = np.asarray(pos, dtype=float)
        x = pos.dot(self.m)
        y = pos.dot(self.n)
        
        # Compute arctan(y/x) for all values
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            theta = np.arctan(y / x)
        
        # Handle special cases and ensure value range -π to π
        theta[(x == 0) & (y > 0)] = np.pi / 2
        theta[(x == 0) & (y < 0)] = -np.pi / 2
        theta[(x < 0)] += np.pi 
        theta[(theta >= np.pi)] -= 2 * np.pi
        
        return theta
    
    def displacement(self, pos):
        """
        Compute the position-dependent isotropic displacements.
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
        
        Returns
        -------
        numpy.ndarray
            The computed 3D vector displacements at all given points.
        """
        pos = np.asarray(pos)
        if pos.shape == (3,):
            pos = pos.reshape(1,3)

        # Split pos, burgers into components
        x = pos.dot(self.m)
        y = pos.dot(self.n)
        b_s = self.burgers.dot(self.ξ)
        b_e = self.burgers.dot(self.m)
        nu = self.nu
        
        # Compute displacement components in m, n, ξ directions
        disp_m = b_e / (2 * np.pi) * (self.theta(pos) + (x * y) / (2 * (1 - nu) * (x**2 + y**2)))
        
        disp_n = b_e / (2 * np.pi) * (-(1 - 2 * nu) / (4 * (1 - nu)) * np.log(x**2 + y**2)
                           + (y**2) / (2 * (1 - nu) * (x**2 + y**2)))
        
        disp_ξ = b_s / (2 * np.pi) * (self.theta(pos))

        # Combine into array
        disp = np.outer(disp_ξ, self.ξ) + np.outer(disp_m, self.m) + np.outer(disp_n, self.n)
        if disp.shape[0] == 1:
            return disp[0]
        else:
            return disp

    def stress(self, pos):
        """
        Compute the position-dependent isotropic stresses.
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
        
        Returns
        -------
        numpy.ndarray
            The computed 3x3 stress states at all given points.
        """
        pos = np.asarray(pos)
        if pos.shape == (3,):
            pos = pos.reshape(1,3)

        # Split pos, burgers into components
        x = pos.dot(self.m)
        y = pos.dot(self.n)
        b_s = self.burgers.dot(self.ξ)
        b_e = self.burgers.dot(self.m)
        nu = self.nu
        mu = self.mu
        
        # Define screw / edge prefactors
        pre_s = mu * b_s / (2 * np.pi)
        pre_e = mu * b_e / (2 * np.pi * (1 - nu))

        # Initialize empty stress array
        sigma = np.empty(pos.shape[:-1] + (3,3))
  
        # Stress components due to b_s
        sigma[..., 0, 2] = sigma[..., 2, 0] =-pre_s * y / (x**2 + y**2)
        sigma[..., 1, 2] = sigma[..., 2, 1] = pre_s * x / (x**2 + y**2)

        # Stress components due to b_e
        sigma[..., 0, 0] =-pre_e * (y * (3 * x**2 + y**2)) / (x**2 + y**2)**2
        sigma[..., 1, 1] = pre_e * (y * (x**2 - y**2)) / (x**2 + y**2)**2
        sigma[..., 2, 2] = nu * (sigma[..., 0, 0] + sigma[..., 1, 1])
        sigma[..., 0, 1] = sigma[..., 1, 0] = pre_e * (x * (x**2 - y**2)) / (x**2 + y**2)**2

        # Get the reverse transformation matrix
        transform = np.array([self.m, self.n, self.ξ]).T

        # Transform stresses
        return np.einsum('mi,nj,...ij->...mn', transform, transform, sigma)



# coding: utf-8
# atomman imports
from atomman.lammps import style

def atoms_prop_info(atom_style='atomic', units='metal'):
    """
    Returns the prop_info associated with an Atoms section in a LAMMPS data file
    
    Parameters
    ----------
    atom_style : str, optional
        The LAMMPS atom_style option.  Default value is 'atomic'.
    units : str, optional
        The LAMMPS units option.  Default value is 'metal'.
        
    Returns
    -------
    prop_info : list of dict
        The proper conversion metadata as used by atomman.convert.atoms_table
    """
    lammps_unit = style.unit(units)
    
    if atom_style == 'angle':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'atomic':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'body':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "bflag",
                      "table_name": "bodyflag"},
                     
                     {"prop_name": "mass",
                      "table_name": "mass",
                      "unit": lammps_unit['mass']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'bond':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'charge':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "charge",
                      "table_name": "q",
                      "unit": lammps_unit['charge']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"], 
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'dipole':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "charge",
                      "table_name": "q",
                      "unit": lammps_unit['charge']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "mu",
                      "table_name": ["mux", "muy", "muz"],
                      "unit": lammps_unit['dipole']}]
    
    elif atom_style == 'electron':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "charge",
                      "table_name": "q",
                      "unit": lammps_unit['charge']},
                     
                     {"prop_name": "espin",
                      "table_name": "spin"},
                     
                     {"prop_name": "eradius",
                      "table_name": "eradius", 
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'ellipsoid':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "eflag",
                      "table_name": "ellipsoidflag"},
                     
                     {"prop_name": "density",
                      "table_name": "density",
                      "unit": lammps_unit['density']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'full':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "charge",
                      "table_name": "q",
                      "unit": lammps_unit['charge']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'line':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "lflag",
                      "table_name": "lineflag"},
                     
                     {"prop_name": "density",
                      "table_name": "density",
                      "unit": lammps_unit['density']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'meso':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "rho",
                      "table_name": "rho"},
                     
                     {"prop_name": "e",
                      "table_name": "e"},
                     
                     {"prop_name": "cv",
                      "table_name": "cv"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'molecular':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'peri':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "volume",
                      "table_name": "volume",
                      "unit": lammps_unit['volume']},
                     
                     {"prop_name": "density",
                      "table_name": "density",
                      "unit": lammps_unit['density']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'smd':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "volume",
                      "table_name": "volume",
                      "unit": lammps_unit['volume']},
                     
                     {"prop_name": "mass",
                      "table_name": "mass",
                      "unit": lammps_unit['mass']},
                     
                     {"prop_name": "kradius",
                      "table_name": "kernalradius",
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "cradius",
                      "table_name": "contactradius",
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'sphere':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "diameter",
                      "table_name": "diameter",
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "density",
                      "table_name": "density",
                      "unit": lammps_unit['density']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'template':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "m_template",
                      "table_name": "templateindex"},
                     
                     {"prop_name": "a_template",
                      "table_name": "templateatom"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'tri':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "m_id",
                      "table_name": "mol"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "tflag",
                      "table_name": "triangleflag"},
                     
                     {"prop_name": "density",
                      "table_name": "density",
                      "unit": lammps_unit['density']},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style == 'wavepacket':
        prop_info = [{"prop_name": "a_id",
                      "table_name": "id"},
                     
                     {"prop_name": "atype",
                      "table_name": "type"},
                     
                     {"prop_name": "charge",
                      "table_name": "q",
                      "unit": lammps_unit['charge']},
                     
                     {"prop_name": "espin",
                      "table_name": "spin"},
                     
                     {"prop_name": "eradius",
                      "table_name": "eradius", 
                      "unit": lammps_unit['length']},
                     
                     {"prop_name": "e_id",
                      "table_name": "etag"},
                     
                     {"prop_name": "cs_re",
                      "table_name": "cs_re"},
                     
                     {"prop_name": "cs_im",
                      "table_name": "cs_im"},
                     
                     {"prop_name": "pos",
                      "table_name": ["x", "y", "z"],
                      "unit": lammps_unit['length']}]
    
    elif atom_style[:6] == 'hybrid':
        substyles = atom_style.split()
        prop_info = atoms_prop_info('atomic')
        for substyle in substyles[1:]:
            prop_names = []
            for prop in prop_info:
                prop_names.append(prop['prop_name'])
            subprop_info = atoms_prop_info(substyle)
            for prop in subprop_info:
                if prop['prop_name'] not in prop_names:
                    prop_info.append(prop)
    else:
        raise ValueError('atom_style ' + atom_style + ' not supported')
    
    return prop_info


# coding: utf-8

def euler(ratefxn, coord, timestep, **kwargs):
    """
    Performs Euler ODE integration for a timestep.
    
    Parameters
    ----------
    ratefxn : function
        The rate function to use.
    coord : array-like object
        The coordinate(s) of the last timestep.
    timestep : float
        The timestep value to use.
    
    Returns
    -------
    array-like object
        The coordinate(s) moved forward by timestep.
    """
    
    return coord + timestep * ratefxn(coord, **kwargs)


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from ... import Atoms, Box, System

def load(cell, symbols=()):
    """
    Convert an spglib cell tuple into an atomman.System
    
    Parameters
    ----------
    cell : tuple
        A tuple containing 3x3 lattice vectors, 3XN box relative positions,
        and N numeric atomic types.
    symbols : list
        The elemental symbols to pair with the unique atom types/numbers.
    
    Returns
    -------
    system : atomman.System
        A atomman representation of a system.
    """
    
    # Separate out cell tuple
    lattice = cell[0]
    positions = cell[1]
    numbers = cell[2]
    
    # Define box
    box = Box(vects = lattice)
    
    # Define atoms
    atoms = Atoms(atype=numbers, pos=positions)
    
    # Build system
    system = System(atoms=atoms, box=box, symbols=symbols, scale=True)
    
    return system


# coding: utf-8

# http://www.numpy.org/
import numpy as np

from . import Shape, Plane

class PlaneSet(Shape):
    """
    Class consisting of a shape defined by a set of planes
    """
    def __init__(self, planes):
        """
        Defines a shape based on a list of Plane objects
        
        Parameters
        ----------
        planes : list of atomman.region.Plane
            The planes to use in constructing the shape.  Points "below" all
            planes are considered inside the shape.
        """
        self.planes = planes
    
    @property 
    def planes(self):
        """list of atomman.region.Plane : The planes that make up the shape"""
        return self.__planes
    
    @planes.setter
    def planes(self, value):
        if isinstance(value, Plane):
            value = [value]
        else:
            value = list(value)
            for v in value:
                assert isinstance(v, Plane)
        self.__planes = value

    def inside(self, pos, inclusive=True):
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is True.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if inside shape
        """
        
        # Ensure pos is a numpy array 
        pos = np.asarray(pos)
        
        # Construct test array with all Trues
        insideplanes = np.ones(len(pos), dtype=bool)
        
        # Loop over each plane and check if pos are below
        for plane in self.planes:
            insideplanes = insideplanes & plane.below(pos, inclusive=inclusive)
        
        return insideplanes